# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from decimal import Decimal
from itertools import chain

from amazon.ion.core import timestamp, TimestampPrecision
from amazon.ion.exceptions import IonException
from amazon.ion.reader import ReadEventType
from amazon.ion.reader_text import reader, _POS_INF, _NEG_INF, _NAN
from amazon.ion.symbols import SymbolToken
from amazon.ion.util import coroutine
from tests import listify, parametrize
from tests.event_aliases import *
from tests.reader_util import ReaderParameter, reader_scaffold, all_top_level_as_one_stream_params, value_iter

_P = ReaderParameter
_ts = timestamp
_tp = TimestampPrecision
_d = Decimal
_st = partial(SymbolToken, sid=None, location=None)


def _sid(sid):
    return SymbolToken(text=None, sid=sid, location=None)

_BAD_GRAMMAR = (
    (b'$ion_1_1 42',),
    (b'$ion_10_1 42',),
    (b'$ion_1_02 42',),
    (b'+1',),
    (b'01',),
    (b'1.23.4',),
    (b'1__0',),
    (b'1_e1',),
    (b'1e_1',),
    (b'1e1_',),
    (b'-infs',),
    (b'+infs',),
    (b'-in',),
    (b'1._0',),
    (b'1_.0',),
    (b'-_1',),
    (b'1_',),
    (b'0_x1',),
    (b'0b_1',),
    (b'1e0-1',),
    (b'1e0e-1',),
    (b'+inf-',),
    (b'null.strings',),
    (b'null.strn',),
    (b'null.flat',),
    (b'null.x',),
    (b'null.',),
    (b'200T',),
    (b'10000T',),
    (b'-2000T',),
    (b'-0001T',),
    (b'00-01T',),
    (b'2000-01',),
    (b'2000-001T',),
    (b'2000--01T',),
    (b'2000-01-123T',),
    (b'2000-01-12T1',),
    (b'2000-01--3T',),
    (b'2007-02-23T20:14:33.Z',),
    (b'2007-02-23T20:14:33.12.3Z',),
    (b'2007-02-23T20:14:33.12+00',),
    (b'1a',),
    (b'foo-',),
    (b'%',),
    (b'n%',),
    (b'"\n"',),
    (b'"\a"',),
    (b'"\\\a"',),
    (b'"a\b"',),
    (b'"\\\n\r"',),
    (b'"\0"',),
    (b"'\n'",),
    (b"'\a'",),
    (b"'\\\a'",),
    (b"'a\b'",),
    (b"'\\\n\r'",),
    (b"'\0'",),
    (b"'''\b'''",),
    (b"'''a\b'''",),
    (b'"\\udbff\\""',),  # Unpaired escaped surrogate.
    (b'"\\udbffabcdef',),  # Unpaired escaped surrogate.
    (b"'''\\udbff'''",),  # Splitting surrogate escapes across long string literal boundaries is illegal per the spec.
    (b'abc://',),
    (b'abc/**/://',),
    (b'{{/**/}}',),
    (b'{{//\n}}',),
    (b'{{/**/"abc"}}',),
    (b'{{"abc"//\n}}',),
    (b'{{\'\'\'abc\'\'\'//\n\'\'\'def\'\'\'}}',),
    (b'{{"\xf6"}}',),
    (b'{{"\n"}}',),
    (b"{{'''\0'''}}",),
    (b"{{'''\\u0000'''}}",),
    (b"{{'''\\u3000'''}}",),
    (b'{{"\\u0000"}}',),
    (b'{{"\\u3000"}}',),
    (b'{{"\\U0001f4a9"}}',),
    (b'{{ abcd} }',),
    (b'{ {abcd}}', e_start_struct()),
    (b'{{\'\' \'foo\'\'\'}}',),
    (b'{{\'"foo"}}',),
    (b'{{abc}de}}',),
    (b'{{"abc"}de}',),
    (b'{{ab}}',),
    (b'{{ab=}}',),
    (b'{{ab=}=}',),
    (b'{{ab===}}',),
    (b'{{====}}',),
    (b'{{abcd====}}',),
    (b'{{abc*}}',),
    (b'{foo:bar/**/baz:zar}', e_start_struct(), e_symbol(value=_st(u'bar'), field_name=_st(u'foo'))),
    (b'{foo:bar/**/baz}', e_start_struct(), e_symbol(value=_st(u'bar'), field_name=_st(u'foo'))),
    (b'[abc 123]', e_start_list(), e_symbol(value=_st(u'abc'))),
    (b'[abc/**/def]', e_start_list(), e_symbol(value=_st(u'abc'))),
    (b'{abc:}', e_start_struct()),
    (b'{abc :}', e_start_struct()),
    (b'{abc : //\n}', e_start_struct()),
    (b'[abc:]', e_start_list()),
    (b'(abc:)', e_start_sexp()),
    (b'[abc::]', e_start_list()),
    (b'(abc::)', e_start_sexp()),
    (b'{abc::}', e_start_struct()),
    (b'[abc ::]', e_start_list()),
    (b'(abc/**/::)', e_start_sexp()),
    (b'{abc//\n::}', e_start_struct()),
    (b'[abc::/**/]', e_start_list()),
    (b'(abc:: )', e_start_sexp()),
    (b'{abc:://\n}', e_start_struct()),
    (b'{foo:abc::}', e_start_struct()),
    (b'{foo:abc::/**/}', e_start_struct()),
    (b'{foo::bar}', e_start_struct()),
    (b'{foo::bar:baz}', e_start_struct()),
    (b'{foo, bar}', e_start_struct()),
    (b'{foo}', e_start_struct()),
    (b'{123}', e_start_struct()),
    (b'{42, 43}', e_start_struct()),
    (b'[abc, , 123]', e_start_list(), e_symbol(value=_st(u'abc'))),
    (b'[\'\'\'abc\'\'\'\'\']', e_start_list()),
    (b'[\'\'\'abc\'\'\'\'foo\']', e_start_list()),
    (b'[\'\'\'abc\'\'\'\'\', def]', e_start_list()),
    (b'{foo:\'\'\'abc\'\'\'\'\'}', e_start_struct()),
    (b'{foo:\'\'\'abc\'\'\'\'\', bar:def}', e_start_struct()),
    (b'[,]', e_start_list()),
    (b'(,)', e_start_sexp()),
    (b'{,}', e_start_struct()),
    (b'{foo:bar, ,}', e_start_struct(), e_symbol(value=_st(u'bar'), field_name=_st(u'foo'))),
    (b'{true:123}', e_start_struct()),
    (b'{false:123}', e_start_struct()),
    (b'{+inf:123}', e_start_struct()),
    (b'{-inf:123}', e_start_struct()),
    (b'{nan:123}', e_start_struct()),
    (b'{nan}', e_start_struct()),
    (b'{null.clob:123}', e_start_struct()),
    (b'{%:123}', e_start_struct()),
    (b'\'\'\'foo\'\'\'/\'\'\'bar\'\'\'',),  # Dangling slash at the top level.
    (b'{{\'\'\'foo\'\'\' \'\'bar\'\'\'}}',),
    (b'{\'\'\'foo\'\'\'/**/\'\'bar\'\'\':baz}', e_start_struct()),  # Missing an opening ' before "bar".
    (b'{\'\'\'foo\'\'\'/**/\'\'\'bar\'\'\'a:baz}', e_start_struct()),  # Character after field name, before colon.
    (b'{\'foo\'a:baz}', e_start_struct()),
    (b'{"foo"a:baz}', e_start_struct()),
    (b'(1..)', e_start_sexp()),
    (b'(1.a)', e_start_sexp()),
    (b'(1.23.)', e_start_sexp()),
    (b'(42/)', e_start_sexp()),
    (b'42/',),
    (b'0/',),
    (b'1.2/',),
    (b'1./',),
    (b'1.2e3/',),
    (b'1.2d3/',),
    (b'2000T/',),
    (b'/ ',),
    (b'/b',),
)

_BAD_VALUE = (
    (b'0000T',),  # Years must be 1..9999
    (b'2000-13T',),  # 2000 didn't have a thirteenth month.
    (b'2015-02-29T',),  # 2015 was not a leap year.
    (b'2000-01-01T24:00Z',),  # Hour is 0..23.
    (b'2000-01-01T00:60Z',),  # Minute is 0..59.
    (b'2000-01-01T00:00:60Z',),  # Second is 0..59.
    (b'2000-01-01T00:00:00.000+24:00',),  # Hour offset is 0..23.
    (b'2000-01-01T00:00:00.000+00:60',),  # Minute offset is 0..59.
    (b'"\\udbff\\u3000"',),  # Malformed surrogate pair (\u3000 is not a low surrogate).
    (b'"\\u3000\\udfff"',),  # Malformed surrogate pair (\u3000 is not a high surrogate).
)

_INCOMPLETE = (
    (b'{',),  # Might be a lob.
    (b'{ ', e_start_struct()),
    (b'[', e_start_list()),
    (b'(', e_start_sexp()),
    (b'[[]', e_start_list(), e_start_list(), e_end_list()),
    (b'(()', e_start_sexp(), e_start_sexp(), e_end_sexp()),
    (b'{foo:{}', e_start_struct(), e_start_struct(field_name=_st(u'foo')), e_end_struct()),
    (b'{foo:bar', e_start_struct(),),
    (b'{foo:bar::', e_start_struct(),),
    (b'{foo:bar,', e_start_struct(), e_symbol(_st(u'bar'), field_name=_st(u'foo'))),
    (b'[[],', e_start_list(), e_start_list(), e_end_list()),
    (b'{foo:{},', e_start_struct(), e_start_struct(field_name=_st(u'foo')), e_end_struct()),
    (b'foo',),  # Might be an annotation.
    (b'\'foo\'',),  # Might be an annotation.
    (b'\'\'\'foo\'\'\'/**/',),  # Might be followed by another triple-quoted string.
    (b'\'\'\'\'',),  # Might be followed by another triple-quoted string.
    (b"'''abc''''def'", e_string(u'abc'),),
    (b'123',),  # Might have more digits.
    (b'42/',),  # The / might start a comment
    (b'0/',),
    (b'1.2/',),
    (b'1./',),
    (b'1.2e3/',),
    (b'1.2d3/',),
    (b'2000T/',),
    (b'-',),
    (b'+',),
    (b'1.2',),
    (b'1.2e',),
    (b'1.2e-',),
    (b'+inf',),  # Might be followed by more characters, making it invalid at the top level.
    (b'-inf',),
    (b'nan',),
    (b'1.2d',),
    (b'1.2d3',),
    (b'1_',),
    (b'0b',),
    (b'0x',),
    (b'2000-01',),
    (b'"abc',),
    (b'false',),  # Might be a symbol with more characters.
    (b'true',),
    (b'null.string',),  # Might be a symbol with more characters.
    (b'/',),
    (b'/*',),
    (b'//',),
    (b'foo:',),
    (b'foo::',),
    (b'foo::bar',),
    (b'foo//\n',),
    (b'{foo', e_start_struct()),
    (b'{{',),
    (b'{{"',),
    (b'(foo-', e_start_sexp(), e_symbol(_st(u'foo'))),
    (b'(-foo', e_start_sexp(), e_symbol(_st(u'-'))),
)

_SKIP = (
    [(e_read(b'123 456 '), e_int(123)), (SKIP, TypeError)],  # Can't skip at top-level.
    [(e_read(b'[]'), e_start_list()), (SKIP, e_end_list()), (NEXT, END)],
    [(e_read(b'{//\n}'), e_start_struct()), (SKIP, e_end_struct()), (NEXT, END)],
    [(e_read(b'(/**/)'), e_start_sexp()), (SKIP, e_end_sexp()), (NEXT, END)],
    [(e_read(b'[a,b,c]'), e_start_list()), (NEXT, e_symbol(_st(u'a'))), (SKIP, e_end_list()), (NEXT, END)],
    [
        (e_read(b'{c:a,d:e::b}'), e_start_struct()),
        (NEXT, e_symbol(_st(u'a'), field_name=_st(u'c'))),
        (SKIP, e_end_struct()), (NEXT, END)
    ],
    [
        (e_read(b'(([{a:b}]))'), e_start_sexp()),
        (NEXT, e_start_sexp()),
        (SKIP, e_end_sexp()),
        (NEXT, e_end_sexp()), (NEXT, END)],
    [
        (e_read(b'['), e_start_list()),
        (SKIP, INC),
        (e_read(b'a,42'), INC),
        (e_read(b',]'), e_end_list()), (NEXT, END)
    ],
    [
        (e_read(b'{'), INC),
        (e_read(b'foo'), e_start_struct()),
        (SKIP, INC),
        (e_read(b':bar,baz:zar}'), e_end_struct()), (NEXT, END)
    ],
    [
        (e_read(b'('), e_start_sexp()),
        (SKIP, INC),
        (e_read(b'a+b'), INC),
        (e_read(b'//\n'), INC),
        (e_read(b')'), e_end_sexp()), (NEXT, END)
    ],
)

_NEXT_ERROR = (NEXT, IonException)
_NEXT_INC = (NEXT, INC)
_NEXT_END = (NEXT, END)


_GOOD_FLUSH = (
    [(e_read(b'0'), INC), (NEXT, e_int(0)), _NEXT_END],
    [(e_read(b'1'), INC), (NEXT, e_int(1)), _NEXT_END],
    [(e_read(b'-0'), INC), (NEXT, e_int(0)), _NEXT_END],
    [(e_read(b'123'), INC), (NEXT, e_int(123)), _NEXT_END],
    [(e_read(b'123.'), INC), (NEXT, e_decimal(_d(123))), _NEXT_END],
    [(e_read(b'1.23e-4'), INC), (NEXT, e_float(1.23e-4)), _NEXT_END],
    [(e_read(b'1.23d+4'), INC), (NEXT, e_decimal(_d(u'1.23e4'))), _NEXT_END],
    [(e_read(b'2000-01-01'), INC), (NEXT, e_timestamp(_ts(2000, 1, 1, precision=_tp.DAY))), _NEXT_END],
    [(e_read(b"a"), INC), (NEXT, e_symbol(_st(u'a'))), _NEXT_END],
    [(e_read(b"'abc'"), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END],
    [(e_read(b"$abc"), INC), (NEXT, e_symbol(_st(u'$abc'))), _NEXT_END],
    [(e_read(b"$"), INC), (NEXT, e_symbol(_st(u'$'))), _NEXT_END],
    [(e_read(b"$10"), INC), (NEXT, e_symbol(_sid(10))), _NEXT_END, (e_read(b'0'), INC), (NEXT, e_int(0)), _NEXT_END],
    [(e_read(b'abc'), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END, (e_read(b'def'), INC),
     (NEXT, e_symbol(_st(u'def'))), _NEXT_END],
    [(e_read(b"''"), INC), (NEXT, e_symbol(_st(u''))), _NEXT_END],
    [(e_read(b"'''abc'''"), INC), (NEXT, e_string(u'abc')), (NEXT, END), (e_read(b"'''def'''"), INC),
     (NEXT, e_string(u'def')), _NEXT_END],
    [(e_read(b"'''abc''''def'"), e_string(u'abc')), _NEXT_INC, (NEXT, e_symbol(_st(u'def'))), _NEXT_END],
    [(e_read(b"'''abc'''''"), INC), (NEXT, e_string(u'abc')), (NEXT, e_symbol(_st(u''))), _NEXT_END],
    [(e_read(b"'''abc'''//\n'def'"), e_string(u'abc')), _NEXT_INC, (NEXT, e_symbol(_st(u'def'))), _NEXT_END],
    [(e_read(b"'''abc'''/**/''"), INC), (NEXT, e_string(u'abc')), (NEXT, e_symbol(_st(u''))), _NEXT_END],
    [(e_read(b"'''abc'''//\n/**/''"), INC), (NEXT, e_string(u'abc')), (NEXT, e_symbol(_st(u''))), _NEXT_END],
    [(e_read(b'null'), INC), (NEXT, e_null()), _NEXT_END],
    [(e_read(b'null.string'), INC), (NEXT, e_string()), _NEXT_END],
    [(e_read(b'+inf'), INC), (NEXT, e_float(_POS_INF)), _NEXT_END],
    [(e_read(b'nan'), INC), (NEXT, e_float(_NAN)), _NEXT_END],
    [(e_read(b'true'), INC), (NEXT, e_bool(True)), _NEXT_END],
    [(e_read(b'//'), INC), _NEXT_END],  # Matches ion-java - termination of line comment with newline not required.
    [(e_read(b'abc//123\n'), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END],
    [(e_read(b"'abc'//123\n"), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END],
    [(e_read(b'abc//123'), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END],
    [(e_read(b"'abc'//123"), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END],
)

_BAD_FLUSH = (
    [(e_read(b'$ion_1_1'), INC), _NEXT_ERROR],
    [(e_read(b'123_'), INC), _NEXT_ERROR],
    [(e_read(b'123e'), INC), _NEXT_ERROR],
    [(e_read(b'123e-'), INC), _NEXT_ERROR],
    [(e_read(b'123d+'), INC), _NEXT_ERROR],
    [(e_read(b'0x'), INC), _NEXT_ERROR],
    [(e_read(b'2000-01-'), INC), _NEXT_ERROR],
    [(e_read(b'"'), INC), _NEXT_ERROR],
    [(e_read(b'/'), INC), _NEXT_ERROR],
    [(e_read(b'abc/'), INC), _NEXT_ERROR],
    [(e_read(b'{/'), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b'/*'), INC), _NEXT_ERROR],
    [(e_read(b'abc/*'), INC), _NEXT_ERROR],
    [(e_read(b'[/*'), e_start_list()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b'(//'), e_start_sexp()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b'(//\n'), e_start_sexp()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b'+in'), INC), _NEXT_ERROR],
    [(e_read(b'null.'), INC), _NEXT_ERROR],
    [(e_read(b'null.str'), INC), _NEXT_ERROR],
    [(e_read(b'"abc'), INC), _NEXT_ERROR],
    [(e_read(b"'abc"), INC), _NEXT_ERROR],
    [(e_read(b"'''abc"), INC), _NEXT_ERROR],
    [(e_read(b"'''abc''''"), INC), _NEXT_ERROR],
    [(e_read(b"{{abc"), INC), _NEXT_ERROR],
    [(e_read(b'{{"abc"'), INC), _NEXT_ERROR],
    [(e_read(b"(abc"), e_start_sexp()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"[abc "), e_start_list()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"['abc' "), e_start_list()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc:def "), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc "), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc: "), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc://"), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc:/**/"), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc:/*"), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc//\n:"), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc/**/:"), e_start_struct()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"(abc "), e_start_sexp()), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"[abc,"), e_start_list()), (NEXT, e_symbol(_st(u'abc'))), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"[abc/**/,"), e_start_list()), (NEXT, e_symbol(_st(u'abc'))), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"[abc,//"), e_start_list()), (NEXT, e_symbol(_st(u'abc'))), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"[abc,/**/"), e_start_list()), (NEXT, e_symbol(_st(u'abc'))), _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"{abc:def,"), e_start_struct()), (NEXT, e_symbol(_st(u'def'), field_name=_st(u'abc'))),
     _NEXT_INC, _NEXT_ERROR],
    [(e_read(b"abc:"), INC), _NEXT_ERROR],
    [(e_read(b"abc/**/:"), INC), _NEXT_ERROR],
    [(e_read(b"abc//\n:"), INC), _NEXT_ERROR],
    [(e_read(b"abc::"), INC), _NEXT_ERROR],
    [(e_read(b"'abc'::"), INC), _NEXT_ERROR],
    [(e_read(b"abc:: //"), INC), _NEXT_ERROR],
    [(e_read(b"'abc' ::/**/"), INC), _NEXT_ERROR],
    [(e_read(b"abc//\n:: "), INC), _NEXT_ERROR],
    [(e_read(b"'abc'/**/::"), INC), _NEXT_ERROR],
    [(e_read(b"abc//\n::/**/"), INC), _NEXT_ERROR],
    [(e_read(b"'abc'/**/:://"), INC), _NEXT_ERROR],
    [(e_read(b'abc'), INC), (NEXT, e_symbol(_st(u'abc'))), _NEXT_END, (e_read(b'::123 '), IonException)],
    [(e_read(b'$10'), INC), (NEXT, e_symbol(_sid(10))), _NEXT_END, (e_read(b'::123 '), IonException)],
)


def _good_container(start, end, *events):
    return (start(),) + events + (end(),)

_good_sexp = partial(_good_container, e_start_sexp, e_end_sexp)
_good_struct = partial(_good_container, e_start_struct, e_end_struct)
_good_list = partial(_good_container, e_start_list, e_end_list)


_GOOD = (
    (b'$ion_1_0 42 ', IVM, e_int(42)),
    (b'$ion_1_0_ 42 ', e_symbol(_st(u'$ion_1_0_')), e_int(42)),
    (b'$ion_1_0a 42 ', e_symbol(_st(u'$ion_1_0a')), e_int(42)),
    (b'$ion_1_ 42 ', e_symbol(_st(u'$ion_1_')), e_int(42)),
    (b'$ion_a_b 42 ', e_symbol(_st(u'$ion_a_b')), e_int(42)),
    (b'$ion_1_b 42 ', e_symbol(_st(u'$ion_1_b')), e_int(42)),
    (b'ann::$ion_1_0 42 ', e_symbol(_st(u'$ion_1_0'), annotations=(_st(u'ann'),)), e_int(42)),
    (b'$ion_1234_1::$ion_1_0 42 ', e_symbol(_st(u'$ion_1_0'), annotations=(_st(u'$ion_1234_1'),)), e_int(42)),
    (b'$ion_1_0::$ion_1234_1 42 ', e_symbol(_st(u'$ion_1234_1'), annotations=(_st(u'$ion_1_0'),)), e_int(42)),
    (b'{$ion_1_0:abc}',) + _good_struct(e_symbol(_st(u'abc'), field_name=_st(u'$ion_1_0'))),
    (b'($ion_1_0)',) + _good_sexp(e_symbol(_st(u'$ion_1_0'))),
    (b'42[]', e_int(42)) + _good_list(),
    (b'\'foo\'123 ', e_symbol(_st(u'foo')), e_int(123)),
    (b'null()', e_null()) + _good_sexp(),
    (b'tru{}', e_symbol(_st(u'tru'))) + _good_struct(),
    (b'{{"foo"}}42{{}}', e_clob(b'foo'), e_int(42), e_blob(b'')),
    (b'+inf"bar"', e_float(_POS_INF), e_string(u'bar')),
    (b'foo\'bar\'"baz"', e_symbol(_st(u'foo')), e_symbol(_st(u'bar')), e_string(u'baz')),
    (b'\'\'\'foo\'\'\'\'\'123 ', e_string(u'foo'), e_symbol(_st(u'')), e_int(123)),
    (b'\'\'\'foo\'\'\'\'abc\'123 ', e_string(u'foo'), e_symbol(_st(u'abc')), e_int(123)),
    (b'[]',) + _good_list(),
    (b'()',) + _good_sexp(),
    (b'{}',) + _good_struct(),
    (b'{/**/}',) + _good_struct(),
    (b'(/**/)',) + _good_sexp(),
    (b'[/**/]',) + _good_list(),
    (b'{//\n}',) + _good_struct(),
    (b'(//\n)',) + _good_sexp(),
    (b'[//\n]',) + _good_list(),
    (b'{/**///\n}',) + _good_struct(),
    (b'(/**///\n)',) + _good_sexp(),
    (b'[/**///\n]',) + _good_list(),
    (b'(foo)',) + _good_sexp(e_symbol(_st(u'foo'))),
    (b'[foo]',) + _good_list(e_symbol(_st(u'foo'))),
    (b'(\'\')',) + _good_sexp(e_symbol(_st(u''))),
    (b'[\'\']',) + _good_list(e_symbol(_st(u''))),
    (b'(\'foo\')',) + _good_sexp(e_symbol(_st(u'foo'))),
    (b'[\'foo\']',) + _good_list(e_symbol(_st(u'foo'))),
    (b'/*foo*///bar\n/*baz*/',),
    (b'/*\\n*///\\u3000\n',),
    (b'\'\'::123 ', e_int(123, annotations=(_st(u''),))),
    (b'{foo:zar::[], bar: (), baz:{}}',) + _good_struct(
        e_start_list(field_name=_st(u'foo'), annotations=(_st(u'zar'),)), e_end_list(),
        e_start_sexp(field_name=_st(u'bar')), e_end_sexp(),
        e_start_struct(field_name=_st(u'baz')), e_end_struct()
    ),
    (b'[[], zar::{}, ()]',) + _good_list(
        e_start_list(), e_end_list(),
        e_start_struct(annotations=(_st(u'zar'),)), e_end_struct(),
        e_start_sexp(), e_end_sexp(),
    ),
    (b'{\'\':bar,}',) + _good_struct(e_symbol(_st(u'bar'), field_name=_st(u''))),
    (b'{\'\':bar}',) + _good_struct(e_symbol(_st(u'bar'), field_name=_st(u''))),
    (b'{\'\'\'foo\'\'\'/**/\'\'\'bar\'\'\':baz}',) + _good_struct(e_symbol(_st(u'baz'), field_name=_st(u'foobar')))
)


_GOOD_UNICODE = (
    (u'{foo:bar}',) + _good_struct(e_symbol(_st(u'bar'), field_name=_st(u'foo'))),
    (u'{foo:"b\xf6\u3000r"}',) + _good_struct(e_string(u'b\xf6\u3000r', field_name=_st(u'foo'))),
    (u'{\'b\xf6\u3000r\':"foo"}',) + _good_struct(e_string(u'foo', field_name=_st(u'b\xf6\u3000r'))),
    (u'\x7b\x7d',) + _good_struct(),
    (u'\u005b\u005d',) + _good_list(),
    (u'\u0028\x29',) + _good_sexp(),
    (u'\u0022\u0061\u0062\u0063\u0022', e_string(u'abc')),
    (u'{foo:"b\xf6\U0001f4a9r"}',) + _good_struct(e_string(u'b\xf6\U0001f4a9r', field_name=_st(u'foo'))),
    (u'{\'b\xf6\U0001f4a9r\':"foo"}',) + _good_struct(e_string(u'foo', field_name=_st(u'b\xf6\U0001f4a9r'))),
    (u'{"b\xf6\U0001f4a9r\":"foo"}',) + _good_struct(e_string(u'foo', field_name=_st(u'b\xf6\U0001f4a9r'))),
    (u'{\'\'\'\xf6\'\'\' \'\'\'\U0001f4a9r\'\'\':"foo"}',) + _good_struct(
        e_string(u'foo', field_name=_st(u'\xf6\U0001f4a9r'))
    ),
    (u'\'b\xf6\U0001f4a9r\'::"foo"', e_string(u'foo', annotations=(_st(u'b\xf6\U0001f4a9r'),))),
    (u'"\t\v\f\'"', e_string(u'\t\v\f\'')),
    (u"'''\t\v\f\"\n\r'''42 ", e_string(u'\t\v\f\"\n\n'), e_int(42))
)

_BAD_UNICODE = (
    (u'\xf6',),  # Not an acceptable identifier symbol.
    (u'r\U0001f4a9',),
    (u'{foo:b\xf6\u3000r}', e_start_struct()),
    (u'{b\xf6\u3000:"foo"}', e_start_struct()),
    (u'{br\U0001f4a9:"foo"}', e_start_struct()),
    (u'{br\U0001f4a9r:"foo"}', e_start_struct()),
    (u'{\'\'\'\xf6\'\'\' \'\'\'\U0001f4a9r\'\'\'a:"foo"}', e_start_struct()),
    (u'b\xf6\U0001f4a9r::"foo"',),
    (u"'''\a'''",),
    (u"{{'''\xf6'''}}",),
    (u'{{"\u3000"}}',),
)

_GOOD_ESCAPES_FROM_UNICODE = (
    (u'"\\xf6"', e_string(u'\xf6')),
    (u'"\\a"', e_string(u'\a')),
    (u'"a\\b"', e_string(u'a\b')),
    (u'"\\r"', e_string(u'\r')),
    (u"'\\xf6'42 ", e_symbol(_st(u'\xf6')), e_int(42)),
    (u"'\\a'42 ", e_symbol(_st(u'\a')), e_int(42)),
    (u"'a\\b'42 ", e_symbol(_st(u'a\b')), e_int(42)),
    (u"'\\r'42 ", e_symbol(_st(u'\r')), e_int(42)),
    (u"'''\\b'''42 ", e_string(u'\b'), e_int(42)),
    (u"'''a\\b'''42 ", e_string(u'a\b'), e_int(42)),
    (u'"\\u3000"', e_string(u'\u3000')),
    (u'"\\udbff\\udfff"', e_string(u'\U0010ffff')),  # Escaped surrogate pair.
    (u'["\\U0001F4a9"]',) + _good_list(e_string(u'\U0001f4a9')),
    (u'"\\t "\'\\\'\'"\\v"', e_string(u'\t '), e_symbol(_st(u'\'')), e_string(u'\v')),
    (u'(\'\\/\')',) + _good_sexp(e_symbol(_st(u'/'))),
    (u'{\'\\f\':foo,"\\?":\'\\\\\'::"\\v\\t"}',) + _good_struct(
        e_symbol(_st(u'foo'), field_name=_st(u'\f')), e_string(u'\v\t', field_name=_st(u'?'), annotations=(_st(u'\\'),))
    ),
    (u'\'\\?\\f\'::\'\\xF6\'::"\\\""', e_string(u'"', annotations=(_st(u'?\f'), _st(u'\xf6')))),
    (u"'''\\\'\\\'\\\''''\"\\\'\"", e_string(u"'''"), e_string(u"'")),
    (u"'''a''\\\'b'''\n'''\\\''''/**/''''\'c'''\"\"", e_string(u"a'''b'''c"), e_string(u'')),
    (u"'''foo''''\\U0001f4a9'42 ", e_string(u'foo'), e_symbol(_st(u'\U0001f4a9')), e_int(42)),
    (u"''''\\\r\n'''42 ", e_string(u"'"), e_int(42)),
    (u'"\\\n"', e_string(u'')),
    (u'"\\\r\n"', e_string(u'')),
    (u'"\\\r"', e_string(u'')),
    (u'"\\\r\\xf6"', e_string(u'\xf6')),
    (u'"\\\rabc"', e_string(u'abc')),
    (u"'\\\r\n'::42 ", e_int(42, annotations=(_st(u''),))),
    (u"{'''\\\rfoo\\\n\r''':bar}",) + _good_struct(e_symbol(_st(u'bar'), field_name=_st(u'foo\n'))),
    (u"{{'''\\x00''''''\\x7e'''}}", e_clob(b'\0~')),
    (u"{{'''\\xff'''}}", e_clob(b'\xff')),
    (u'{{"\\t"}}', e_clob(b'\t')),
    (u'{{"\\\n"}}', e_clob(b'')),
    (u"{{'''\\\r\n'''}}", e_clob(b'')),
)

_GOOD_ESCAPES_FROM_BYTES = (
    (br'"\xf6"', e_string(u'\xf6')),
    (br'"\a"', e_string(u'\a')),
    (br'"a\b"', e_string(u'a\b')),
    (br'"\r"', e_string(u'\r')),
    (br"'\xf6'42 ", e_symbol(_st(u'\xf6')), e_int(42)),
    (br"'\a'42 ", e_symbol(_st(u'\a')), e_int(42)),
    (br"'a\b'42 ", e_symbol(_st(u'a\b')), e_int(42)),
    (br"'\r'42 ", e_symbol(_st(u'\r')), e_int(42)),
    (br"'''\b'''42 ", e_string(u'\b'), e_int(42)),
    (br"'''a\b'''42 ", e_string(u'a\b'), e_int(42)),
    (br'"\u3000"', e_string(u'\u3000')),
    (br'"\udbff\udfff"', e_string(u'\U0010ffff')),  # Escaped surrogate pair.
    (br'["\U0001F4a9"]',) + _good_list(e_string(u'\U0001f4a9')),
    (b'"\\t "\'\\\'\'"\\v"', e_string(u'\t '), e_symbol(_st(u'\'')), e_string(u'\v')),
    (b'(\'\\/\')',) + _good_sexp(e_symbol(_st(u'/'))),
    (b'{\'\\f\':foo,"\\?":\'\\\\\'::"\\v\\t"}',) + _good_struct(
        e_symbol(_st(u'foo'), field_name=_st(u'\f')), e_string(u'\v\t', field_name=_st(u'?'), annotations=(_st(u'\\'),))
    ),
    (b'\'\\?\\f\'::\'\\xF6\'::"\\\""', e_string(u'"', annotations=(_st(u'?\f'), _st(u'\xf6')))),
    (b"'''\\\'\\\'\\\''''\"\\\'\"", e_string(u"'''"), e_string(u"'")),
    (b"'''a''\\\'b'''\n'''\\\''''/**/''''\'c'''\"\"", e_string(u"a'''b'''c"), e_string(u'')),
    (b"'''foo''''\\U0001f4a9'42 ", e_string(u'foo'), e_symbol(_st(u'\U0001f4a9')), e_int(42)),
    (b"''''\\\r\n'''42 ", e_string(u"'"), e_int(42)),
    (b'"\\\n"', e_string(u'')),
    (b'"\\\r\n"', e_string(u'')),
    (b'"\\\r"', e_string(u'')),
    (b'"\\\r\\xf6"', e_string(u'\xf6')),
    (b'"\\\rabc"', e_string(u'abc')),
    (b"'\\\r\n'::42 ", e_int(42, annotations=(_st(u''),))),
    (b"{'''\\\rfoo\\\n\r''':bar}",) + _good_struct(e_symbol(_st(u'bar'), field_name=_st(u'foo\n'))),
    (b"{{'''\\x00''''''\\x7e'''}}", e_clob(b'\0~')),
    (b"{{'''\\xff'''}}", e_clob(b'\xff')),
    (b'{{"\\t"}}', e_clob(b'\t')),
    (b'{{"\\\n"}}', e_clob(b'')),
    (b"{{'''\\\r\n'''}}", e_clob(b'')),
)

_INCOMPLETE_ESCAPES = (
    [(e_read(u'"\\'), INC), (e_read(u't"'), e_string(u'\t')), (NEXT, END)],
    [
        (e_read(u'\'\\x'), INC), (e_read(u'f'), INC), (e_read(u'6\'42 '), e_symbol(_st(u'\xf6'))),
        (NEXT, e_int(42)), (NEXT, END)
    ],
    [
        (e_read(u'{"\\U0001'), e_start_struct()), (NEXT, INC), (e_read(u'f4a9"'), INC),
        (e_read(u':bar}'), e_symbol(_st(u'bar'), field_name=_st(u'\U0001f4a9'))), (NEXT, e_end_struct()), (NEXT, END)
    ],
    [
        (e_read(u"'''\\\r"), INC), (e_read(u"\n'''//\n"), INC), (e_read(u"'''abc'''42 "), e_string(u'abc')),
        (NEXT, e_int(42)), (NEXT, END)
    ],
)

_UNICODE_SURROGATES = (
    # Note: Surrogates only allowed with UCS2.
    [(e_read(u'"\ud83d\udca9"'), e_string(u'\U0001f4a9')), (NEXT, END)],
    [(e_read(u'"\ud83d'), INC), (e_read(u'\udca9"'), e_string(u'\U0001f4a9')), (NEXT, END)],
)

_BAD_ESCAPES_FROM_UNICODE = (
    (u'"\\g"',),
    (u'\'\\q\'',),
    (u'\\t',),
    (u'"abc"\\t', e_string(u'abc')),
    (u'\'abc\'\\n', e_symbol(_st(u'abc'))),
    (u'\'abc\'\\xf6', e_symbol(_st(u'abc'))),
    (u"'''abc'''\\U0001f4a9", e_string(u'abc')),
    (u"''\\u3000", e_symbol(_st(u''))),
    (u"'''\\u3''' '''000'''42 ",),
    (u'"\\U0001f4aQ"',),
    (u"{{'''abc'''\\n}}",),
    (u'{{"abc"\\n}}',),
    (u'{\'foo\'\\v:bar}', e_start_struct()),
    (u'{\'\'\'foo\'\'\'\\xf6:bar}', e_start_struct()),
)

_BAD_ESCAPES_FROM_BYTES = (
    (br'"\g"',),
    (br'\'\q\'',),
    (b'\\t',),
    (b'"abc"\\t', e_string(u'abc')),
    (b'\'abc\'\\n', e_symbol(_st(u'abc'))),
    (b'\'abc\'\\xf6', e_symbol(_st(u'abc'))),
    (b"'''abc'''\\U0001f4a9", e_string(u'abc')),
    (b"''\\u3000", e_symbol(_st(u''))),
    (b"'''\\u3''' '''000'''42 ",),
    (b'"\\U0001f4aQ"',),
    (b"{{'''abc'''\\n}}",),
    (b'{{"abc"\\n}}',),
    (b'{\'foo\'\\v:bar}', e_start_struct()),
    (b'{\'\'\'foo\'\'\'\\xf6:bar}', e_start_struct()),
)

_UNSPACED_SEXPS = (
    (b'(a/b)',) + _good_sexp(e_symbol(_st(u'a')), e_symbol(_st(u'/')), e_symbol(_st(u'b'))),
    (b'(a+b)',) + _good_sexp(e_symbol(_st(u'a')), e_symbol(_st(u'+')), e_symbol(_st(u'b'))),
    (b'(a-b)',) + _good_sexp(e_symbol(_st(u'a')), e_symbol(_st(u'-')), e_symbol(_st(u'b'))),
    (b'(/%)',) + _good_sexp(e_symbol(_st(u'/%'))),
    (b'(foo //bar\n::baz)',) + _good_sexp(e_symbol(_st(u'baz'), annotations=(_st(u'foo'),))),
    (b'(foo/*bar*/ ::baz)',) + _good_sexp(e_symbol(_st(u'baz'), annotations=(_st(u'foo'),))),
    (b'(\'a b\' //\n::cd)',) + _good_sexp(e_symbol(_st(u'cd'), annotations=(_st(u'a b'),))),
    (b'(abc//baz\n-)',) + _good_sexp(e_symbol(_st(u'abc')), e_symbol(_st(u'-'))),
    (b'(null-100/**/)',) + _good_sexp(e_null(), e_int(-100)),
    (b'(//\nnull//\n)',) + _good_sexp(e_null()),
    (b'(abc/*baz*/123)',) + _good_sexp(e_symbol(_st(u'abc')), e_int(123)),
    (b'(abc/*baz*/-)',) + _good_sexp(e_symbol(_st(u'abc')), e_symbol(_st(u'-'))),
    (b'(abc//baz\n123)',) + _good_sexp(e_symbol(_st(u'abc')), e_int(123)),
    (b'(abc//\n/123)',) + _good_sexp(e_symbol(_st(u'abc')), e_symbol(_st(u'/')), e_int(123)),
    (b'(abc/////\n/123)',) + _good_sexp(e_symbol(_st(u'abc')), e_symbol(_st(u'/')), e_int(123)),
    (b'(abc/**//123)',) + _good_sexp(e_symbol(_st(u'abc')), e_symbol(_st(u'/')), e_int(123)),
    (b'(foo%+null-//\n)',) + _good_sexp(
        e_symbol(_st(u'foo')), e_symbol(_st(u'%+')), e_null(), e_symbol(_st(u'-//'))  # Matches java.
    ),
    (b'(null-100)',) + _good_sexp(e_null(), e_int(-100)),
    (b'(null\'a\')',) + _good_sexp(e_null(), e_symbol(_st(u'a'))),
    (b'(null\'a\'::b)',) + _good_sexp(e_null(), e_symbol(_st(u'b'), annotations=(_st(u'a'),))),
    (b'(null.string.b)',) + _good_sexp(e_string(None), e_symbol(_st(u'.')), e_symbol(_st(u'b'))),
    (b'(\'\'\'abc\'\'\'\'\')',) + _good_sexp(e_string(u'abc'), e_symbol(_st(u''))),
    (b'(\'\'\'abc\'\'\'\'foo\')',) + _good_sexp(e_string(u'abc'), e_symbol(_st(u'foo'))),
    (b'(\'\'\'abc\'\'\'\'\'42)',) + _good_sexp(e_string(u'abc'), e_symbol(_st(u'')), e_int(42)),
    (b'(42\'a\'::b)',) + _good_sexp(e_int(42), e_symbol(_st(u'b'), annotations=(_st(u'a'),))),
    (b'(1.23[])',) + _good_sexp(e_decimal(_d(u'1.23')), e_start_list(), e_end_list()),
    (b'(\'\'\'foo\'\'\'/\'\'\'bar\'\'\')',) + _good_sexp(e_string(u'foo'), e_symbol(_st(u'/')), e_string(u'bar')),
    (b'(-100)',) + _good_sexp(e_int(-100)),
    (b'(-1.23 .)',) + _good_sexp(e_decimal(_d(u'-1.23')), e_symbol(_st(u'.'))),
    (b'(1.)',) + _good_sexp(e_decimal(_d(u'1.'))),
    (b'(1. .1)',) + _good_sexp(e_decimal(_d(u'1.')), e_symbol(_st(u'.')), e_int(1)),
    (b'(2001-01-01/**/a)',) + _good_sexp(e_timestamp(_ts(2001, 1, 1, precision=_tp.DAY)), e_symbol(_st(u'a'))),
    (b'(nul)',) + _good_sexp(e_symbol(_st(u'nul'))),
    (b'(foo::%-bar)',) + _good_sexp(e_symbol(_st(u'%-'), annotations=(_st(u'foo'),)), e_symbol(_st(u'bar'))),
    (b'(true.False+)',) + _good_sexp(e_bool(True), e_symbol(_st(u'.')), e_symbol(_st(u'False')), e_symbol(_st(u'+'))),
    (b'(false)',) + _good_sexp(e_bool(False)),
    (b'(-inf)',) + _good_sexp(e_float(_NEG_INF)),
    (b'(+inf)',) + _good_sexp(e_float(_POS_INF)),
    (b'(nan)',) + _good_sexp(e_float(_NAN)),
    (b'(-inf+inf)',) + _good_sexp(e_float(_NEG_INF), e_float(_POS_INF)),
    (b'(+inf\'foo\')',) + _good_sexp(e_float(_POS_INF), e_symbol(_st(u'foo'))),
    (b'(-inf\'foo\'::bar)',) + _good_sexp(e_float(_NEG_INF), e_symbol(_st(u'bar'), annotations=(_st(u'foo'),))),
    # TODO the inf tests do not match ion-java's behavior. They should be reconciled. I believe this is more correct.
    (b'(- -inf-inf-in-infs-)',) + _good_sexp(
        e_symbol(_st(u'-')), e_float(_NEG_INF), e_float(_NEG_INF), e_symbol(_st(u'-')),
        e_symbol(_st(u'in')), e_symbol(_st(u'-')), e_symbol(_st(u'infs')), e_symbol(_st(u'-'))
    ),
    (b'(+ +inf+inf+in+infs+)',) + _good_sexp(
        e_symbol(_st(u'+')), e_float(_POS_INF), e_float(_POS_INF), e_symbol(_st(u'+')),
        e_symbol(_st(u'in')), e_symbol(_st(u'+')), e_symbol(_st(u'infs')), e_symbol(_st(u'+'))
    ),
    (b'(nan-nan+nan)',) + _good_sexp(
        e_float(_NAN), e_symbol(_st(u'-')), e_float(_NAN), e_symbol(_st(u'+')),
        e_float(_NAN)
    ),
    (b'(nans-inf+na-)',) + _good_sexp(
        e_symbol(_st(u'nans')), e_float(_NEG_INF), e_symbol(_st(u'+')),
        e_symbol(_st(u'na')), e_symbol(_st(u'-'))
    ),
    (b'({}()zar::[])',) + _good_sexp(
        e_start_struct(), e_end_struct(),
        e_start_sexp(), e_end_sexp(),
        e_start_list(annotations=(_st(u'zar'),)), e_end_list()
    ),
)

_GOOD_SCALARS = (
    (b'null', e_null()),

    (b'false', e_bool(False)),
    (b'true', e_bool(True)),
    (b'null.bool', e_bool()),

    (b'null.int', e_int()),
    (b'0', e_int(0)),
    (b'1_2_3', e_int(123)),
    (b'0xfe', e_int(254)),
    (b'0b101', e_int(5)),
    (b'0b10_1', e_int(5)),
    (b'-0b101', e_int(-5)),
    (b'-0b10_1', e_int(-5)),
    (b'1', e_int(1)),
    (b'-1', e_int(-1)),
    (b'0xc1c2', e_int(49602)),
    (b'0xc1_c2', e_int(49602)),
    (b'-0xc1c2', e_int(-49602)),
    (b'-0xc1_c2', e_int(-49602)),
    (b'9223372036854775808', e_int(9223372036854775808)),
    (b'-9223372036854775809', e_int(-9223372036854775809)),

    (b'null.float', e_float()),
    (b'0.0e1', e_float(0.)),
    (b'-0.0e-1', e_float(-0.)),
    (b'0.0e+1', e_float(0.)),
    (b'0.0E1', e_float(0.)),
    (b'-inf', e_float(_NEG_INF)),
    (b'+inf', e_float(_POS_INF)),
    (b'nan', e_float(_NAN)),

    (b'null.decimal', e_decimal()),
    (b'0.0', e_decimal(_d(u'0.0'))),
    (b'0.', e_decimal(_d(u'0.'))),
    (b'-0.0', e_decimal(_d(u'-0.0'))),
    (b'0d-1000', e_decimal(_d(u'0e-1000'))),
    (b'0d1000', e_decimal(_d(u'0e1000'))),
    (b'1d1', e_decimal(_d(u'1e1'))),
    (b'1D1', e_decimal(_d(u'1e1'))),
    (b'1234d-20', e_decimal(_d(u'1234e-20'))),
    (b'1234d+20', e_decimal(_d(u'1234e20'))),
    (b'1d0', e_decimal(_d(u'1e0'))),
    (b'1d-1', e_decimal(_d(u'1e-1'))),
    (b'0d-1', e_decimal(_d(u'0e-1'))),
    (b'0d1', e_decimal(_d(u'0e1'))),
    (b'-1d1', e_decimal(_d(u'-1e1'))),
    (b'-1d0', e_decimal(_d(u'-1e0'))),
    (b'-1d-1', e_decimal(_d(u'-1e-1'))),
    (b'-0d-1', e_decimal(_d(u'-0e-1'))),
    (b'-0d1', e_decimal(_d(u'-0e1'))),

    (b'null.timestamp', e_timestamp()),
    (b'2007-01T', e_timestamp(_ts(2007, 1, precision=_tp.MONTH))),
    (b'2007T', e_timestamp(_ts(2007, precision=_tp.YEAR))),
    (b'2007-01-01', e_timestamp(_ts(2007, 1, 1, precision=_tp.DAY))),
    (
        b'2000-01-01T00:00:00.0Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, 0, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=1
        ))
    ),
    (
        b'2000-01-01T00:00:00.000Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, 0, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=3
        ))
    ),
    (
        b'2000-01-01T00:00:00.999999Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, 999999, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=6
        ))
    ),
    (
        b'2000-01-01T00:00:00.99999900000Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, 999999, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=6
        ))
    ),
    (
        b'2000-01-01T00:00:00.9999999Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, None, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=None,
            fractional_seconds=Decimal('0.9999999')
        ))
    ),
    (
        b'2000-01-01T00:00:00.1234567Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, None, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=None,
            fractional_seconds=Decimal('0.1234567')
        ))
    ),
    (
        b'2000-01-01T00:00:00.1234567800Z',
        e_timestamp(_ts(
            2000, 1, 1, 0, 0, 0, None, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=None,
            fractional_seconds=Decimal('0.1234567800')
        ))
    ),
    (
        b'2000-01-01T00:00:00.000-00:00',
        e_timestamp(_ts(2000, 1, 1, 0, 0, 0, 0, precision=_tp.SECOND, fractional_precision=3))
    ),
    (
        b'2007-02-23T00:00+00:00',
        e_timestamp(_ts(2007, 2, 23, 0, 0, off_hours=0, off_minutes=0, precision=_tp.MINUTE))
    ),
    (b'2007-01-01T', e_timestamp(_ts(2007, 1, 1, precision=_tp.DAY))),
    (b'2000-01-01T00:00:00Z', e_timestamp(_ts(2000, 1, 1, 0, 0, 0, off_hours=0, off_minutes=0, precision=_tp.SECOND))),
    (
        b'2007-02-23T00:00:00-00:00',
        e_timestamp(_ts(2007, 2, 23, 0, 0, 0, precision=_tp.SECOND))
    ),
    (
        b'2007-02-23T12:14:33.079-08:00',
        e_timestamp(_ts(
            2007, 2, 23, 12, 14, 33, 79000, off_hours=-8, off_minutes=0, precision=_tp.SECOND, fractional_precision=3
        ))
    ),
    (
        b'2007-02-23T20:14:33.079Z',
        e_timestamp(_ts(
            2007, 2, 23, 20, 14, 33, 79000, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=3
        ))
    ),
    (
        b'2007-02-23T20:14:33.079+00:00',
        e_timestamp(_ts(
            2007, 2, 23, 20, 14, 33, 79000, off_hours=0, off_minutes=0, precision=_tp.SECOND, fractional_precision=3
        ))
    ),
    (b'0001T', e_timestamp(_ts(1, precision=_tp.YEAR))),
    (b'0001-01-01T00:00:00Z', e_timestamp(_ts(1, 1, 1, 0, 0, 0, off_hours=0, off_minutes=0, precision=_tp.SECOND))),
    (b'2016-02-29T', e_timestamp(_ts(2016, 2, 29, precision=_tp.DAY))),

    (b'null.symbol', e_symbol()),
    (b'nul', e_symbol(_st(u'nul'))),  # See the logic in the event generators that forces these to emit an event.
    (b'$foo', e_symbol(_st(u'$foo'))),
    (b'$10', e_symbol(_sid(10))),
    (b'$10n', e_symbol(_st(u'$10n'))),
    (b'$2', e_symbol(_sid(2))),  # Note: NOT an IVM event
    (b"'$ion_1_0'", e_symbol(_st(u'$ion_1_0'))),  # Note: NOT an IVM event
    (b'$', e_symbol(_st(u'$'))),
    (b'\'a b\'', e_symbol(_st(u'a b'))),
    (b'\'\'', e_symbol(_st(u''))),

    (b'null.string', e_string()),
    (b'" "', e_string(u' ')),
    (b'\'\'\'foo\'\'\' \'\'\'\'\'\' \'\'\'""\'\'\'', e_string(u'foo""')),
    (b'\'\'\'ab\'\'cd\'\'\'', e_string(u'ab\'\'cd')),
    (b"'''\r\n \r \n \n\r'''", e_string(u'\n \n \n \n\n')),

    (b'null.clob', e_clob()),
    (b'{{""}}', e_clob(b'')),
    (b'{{ "abcd" }}', e_clob(b'abcd')),
    (b'{{"abcd"}}', e_clob(b'abcd')),
    (b'{{"abcd"\n}}', e_clob(b'abcd')),
    (b'{{\'\'\'ab\'\'\' \'\'\'cd\'\'\'}}', e_clob(b'abcd')),
    (b'{{\'\'\'ab\'\'\'\n\'\'\'cd\'\'\'}}', e_clob(b'abcd')),

    (b'null.blob', e_blob()),
    (b'{{}}', e_blob(b'')),
    (b'{{ YW1heg== }}', e_blob(b'amaz')),
    (b'{{ YW1hem8= }}', e_blob(b'amazo')),
    (b'{{ YW1hem9u }}', e_blob(b'amazon')),
    (b'{{ YW1heg = = }}', e_blob(b'amaz')),
    (b'{{aW\n9u}}', e_blob(b'ion')),
    (b'{{aW9u}}', e_blob(b'ion')),

    (b'null.list', e_null_list()),

    (b'null.sexp', e_null_sexp()),

    (b'null.struct', e_null_struct()),
)


def _scalar_event_pairs(data, events, info):
    """Generates event pairs for all scalars.

    Each scalar is represented by a sequence whose first element is the raw data and whose following elements are the
    expected output events.
    """
    first = True
    delimiter, in_container = info
    space_delimited = not (b',' in delimiter)
    for event in events:
        input_event = NEXT
        if first:
            input_event = e_read(data + delimiter)
            if space_delimited and event.value is not None \
                and ((event.ion_type is IonType.SYMBOL) or
                     (event.ion_type is IonType.STRING and
                      b'"'[0] != data[0])):  # triple-quoted strings
                # Because annotations and field names are symbols, a space delimiter after a symbol isn't enough to
                # generate a symbol event immediately. Similarly, triple-quoted strings may be followed by another
                # triple-quoted string if only delimited by whitespace or comments.
                yield input_event, INC
                if in_container:
                    # Within s-expressions, these types are delimited in these tests by another value - in this case,
                    # int 0 (but it could be anything).
                    yield e_read(b'0' + delimiter), event
                    input_event, event = (NEXT, e_int(0))
                else:
                    # This is a top-level value, so it may be flushed with NEXT after INCOMPLETE.
                    input_event, event = (NEXT, event)
            first = False
        yield input_event, event


_scalar_iter = partial(value_iter, _scalar_event_pairs, _GOOD_SCALARS)


@coroutine
def _scalar_params():
    """Generates scalars as reader parameters."""
    while True:
        info = yield
        for data, event_pairs in _scalar_iter(info):
            yield _P(
                desc=data,
                event_pairs=event_pairs + [(NEXT, INC)]
            )


def _top_level_value_params(delimiter=b' ', is_delegate=False):
    """Converts the top-level tuple list into parameters with appropriate ``NEXT`` inputs.

    The expectation is starting from an end of stream top-level context.
    """
    info = (delimiter, False)
    for data, event_pairs in _scalar_iter(info):
        _, first = event_pairs[0]
        if first.event_type is IonEventType.INCOMPLETE:  # Happens with space-delimited symbol values.
            _, first = event_pairs[1]
        yield _P(
            desc='TL %s - %s - %r' %
                 (first.event_type.name, first.ion_type.name, data),
            event_pairs=[(NEXT, END)] + event_pairs + [(NEXT, END)],
        )
    if is_delegate:
        yield


@coroutine
def _all_scalars_in_one_container_params():
    """Generates one parameter that contains all scalar events in a single container. """
    while True:
        info = yield

        @listify
        def generate_event_pairs():
            for data, event_pairs in _scalar_iter(info):
                pairs = ((i, o) for i, o in event_pairs)
                while True:
                    try:
                        input_event, output_event = next(pairs)
                        yield input_event, output_event
                        if output_event is INC:
                            # This is a symbol value.
                            yield next(pairs)  # Input: a scalar. Output: the symbol value's event.
                            yield next(pairs)  # Input: NEXT. Output: the previous scalar's event.
                        yield (NEXT, INC)
                    except StopIteration:
                        break

        yield _P(
            desc='ALL',
            event_pairs=generate_event_pairs()
        )


def _collect_params(param_generator, info):
    """Collects all output of the given coroutine into a single list."""
    params = []
    while True:
        param = param_generator.send(info)
        if param is None:
            return params
        params.append(param)


_TEST_SYMBOLS = (
    (
        b'foo',
        b'$foo',
        b'$ios',
        b'$',
        b'$10',
        b'\'a b\'',
        b'foo ',
        b'\'a b\' ',
        b'foo/*bar*/',
        b'\'a b\' //bar\r',
        b'\'\'',
        b'\'\\U0001f4a9\'',
    ),
    (
        _st(u'foo'),
        _st(u'$foo'),
        _st(u'$ios'),
        _st(u'$'),
        _sid(10),
        _st(u'a b'),
        _st(u'foo'),
        _st(u'a b'),
        _st(u'foo'),
        _st(u'a b'),
        _st(u''),
        _st(u'\U0001f4a9'),
    )
)

_TEST_FIELD_NAMES = (
    _TEST_SYMBOLS[0] +
    (
        b'"foo"',
        b'"foo"//bar\n',
        b'\'\'\'foo\'\'\'/*bar*/\'\'\'baz\'\'\'',
        b'//zar\n\'\'\'foo\'\'\'/*bar*/\'\'\'baz\'\'\'',
        b'\'\'\'a \'\'\'\t\'\'\'b\'\'\'',
        b'\'\'\'a \'\'\'\'\'\'b\'\'\'/*zar*/',
        b'\'\'\'\'\'\'',
        b'"\\xf6"',
        b"'''\r\n \r \n \n\r'''",
    ),
    _TEST_SYMBOLS[1] +
    (
        _st(u'foo'),
        _st(u'foo'),
        _st(u'foobaz'),
        _st(u'foobaz'),
        _st(u'a b'),
        _st(u'a b'),
        _st(u''),
        _st(u'\xf6'),
        _st(u'\n \n \n \n\n'),
    )
)


def _generate_annotations():
    """Circularly generates annotations."""
    assert len(_TEST_SYMBOLS[0]) == len(_TEST_SYMBOLS[1])
    i = 1
    num_symbols = len(_TEST_SYMBOLS[0])
    while True:
        yield _TEST_SYMBOLS[0][0:i], _TEST_SYMBOLS[1][0:i]
        i += 1
        if i == num_symbols:
            i = 0


_annotations_generator = _generate_annotations()


@coroutine
def _annotate_params(params, is_delegate=False):
    """Adds annotation wrappers for a given iterator of parameters."""

    while True:
        info = yield
        params_list = _collect_params(params, info)
        test_annotations, expected_annotations = next(_annotations_generator)
        for param in params_list:
            @listify
            def annotated():
                pairs = ((i, o) for i, o in param.event_pairs)
                while True:
                    try:
                        input_event, output_event = next(pairs)
                        if input_event.type is ReadEventType.DATA:
                            data = b''
                            for test_annotation in test_annotations:
                                data += test_annotation + b'::'
                            data += input_event.data
                            input_event = read_data_event(data)
                            if output_event is INC:
                                yield input_event, output_event
                                input_event, output_event = next(pairs)
                            output_event = output_event.derive_annotations(expected_annotations)
                        yield input_event, output_event
                    except StopIteration:
                        break

            yield _P(
                desc='ANN %r on %s' % (expected_annotations, param.desc),
                event_pairs=annotated(),
            )
        if not is_delegate:
            break


def _generate_field_name():
    """Circularly generates field names."""
    assert len(_TEST_FIELD_NAMES[0]) == len(_TEST_FIELD_NAMES[1])
    i = 0
    num_symbols = len(_TEST_FIELD_NAMES[0])
    while True:
        yield _TEST_FIELD_NAMES[0][i], _TEST_FIELD_NAMES[1][i]
        i += 1
        if i == num_symbols:
            i = 0


_field_name_generator = _generate_field_name()


@coroutine
def _containerize_params(param_generator, with_skip=True, is_delegate=False, top_level=True):
    """Adds container wrappers for a given iteration of parameters."""
    while True:
        yield
        for info in ((IonType.LIST, b'[', b']', b','),
                     (IonType.SEXP, b'(', b')', b' '),  # Sexps without delimiters are tested separately
                     (IonType.STRUCT, b'{ ', b'}', b','),  # Space after opening bracket for instant event.
                     (IonType.LIST, b'[/**/', b'//\n]', b'//\r,'),
                     (IonType.SEXP, b'(//\n', b'/**/)', b'/**/'),
                     (IonType.STRUCT, b'{/**/', b'//\r}', b'/**/,')):
            ion_type = info[0]
            params = _collect_params(param_generator, (info[3], True))
            for param in params:
                @listify
                def add_field_names(event_pairs):
                    container = False
                    first = True
                    for read_event, ion_event in event_pairs:
                        if not container and read_event.type is ReadEventType.DATA:
                            field_name, expected_field_name = next(_field_name_generator)
                            data = field_name + b':' + read_event.data
                            read_event = read_data_event(data)
                            ion_event = ion_event.derive_field_name(expected_field_name)
                        if first and ion_event.event_type is IonEventType.CONTAINER_START:
                            # For containers within a struct--only the CONTAINER_START event gets adorned with a
                            # field name
                            container = True
                        first = False
                        yield read_event, ion_event
                start = []
                end = [(e_read(info[2]), e_end(ion_type))]
                if top_level:
                    start = [(NEXT, END)]
                    end += [(NEXT, END)]
                else:
                    end += [(NEXT, INC)]
                start += [
                    (e_read(info[1]), e_start(ion_type)),
                    (NEXT, INC)
                ]
                if ion_type is IonType.STRUCT:
                    mid = add_field_names(param.event_pairs)
                else:
                    mid = param.event_pairs
                desc = 'CONTAINER %s - %s' % (ion_type.name, param.desc)
                yield _P(
                    desc=desc,
                    event_pairs=start + mid + end,
                )
                if with_skip:
                    @listify
                    def only_data_inc(event_pairs):
                        for read_event, ion_event in event_pairs:
                            if read_event.type is ReadEventType.DATA:
                                yield read_event, INC

                    start = start[:-1] + [(SKIP, INC)]
                    mid = only_data_inc(mid)
                    yield _P(
                        desc='SKIP %s' % desc,
                        event_pairs=start + mid + end,
                    )
        if not is_delegate:
            break


def _expect_event(expected_event, data, events, delimiter):
    """Generates event pairs for a stream that ends in an expected event (or exception), given the text and the output
    events preceding the expected event.
    """
    events += (expected_event,)
    outputs = events[1:]
    event_pairs = [(e_read(data + delimiter), events[0])] + list(zip([NEXT] * len(outputs), outputs))
    return event_pairs


@coroutine
def _basic_params(event_func, desc, delimiter, data_event_pairs, is_delegate=False, top_level=True):
    """Generates parameters from a sequence whose first element is the raw data and the following
    elements are the expected output events.
    """
    while True:
        yield
        params = list(zip(*list(value_iter(event_func, data_event_pairs, delimiter))))[1]
        for param in _paired_params(params, desc, top_level):
            yield param
        if not is_delegate:
            break


def _paired_params(params, desc, top_level=True):
    """Generates reader parameters from sequences of input/output event pairs."""
    for event_pairs in params:
        data = event_pairs[0][0].data
        if top_level:
            event_pairs = [(NEXT, END)] + event_pairs
        yield _P(
            desc='%s %s' % (desc, data),
            event_pairs=event_pairs,
            is_unicode=isinstance(data, str)
        )


_ion_exception = partial(_expect_event, IonException)
_bad_grammar_params = partial(_basic_params, _ion_exception, 'BAD GRAMMAR', b' ')
_bad_unicode_params = partial(_basic_params, _ion_exception, 'BAD GRAMMAR - UNICODE', u' ')
_value_error = partial(_expect_event, ValueError)
_bad_value_params = partial(_basic_params, _value_error, 'BAD VALUE', b' ')
_incomplete = partial(_expect_event, INC)
_incomplete_params = partial(_basic_params, _incomplete, 'INC', b'')
_end = partial(_expect_event, END)
_good_params = partial(_basic_params, _end, 'GOOD', b'')
_good_unicode_params = partial(_basic_params, _end, 'GOOD - UNICODE', u'')


@parametrize(*chain(
    _good_params(_GOOD),
    _bad_grammar_params(_BAD_GRAMMAR),
    _bad_value_params(_BAD_VALUE),
    _incomplete_params(_INCOMPLETE),
    _good_unicode_params(_GOOD_UNICODE),
    _good_unicode_params(_GOOD_ESCAPES_FROM_UNICODE),
    _good_params(_GOOD_ESCAPES_FROM_BYTES),
    _bad_unicode_params(_BAD_UNICODE),
    _bad_unicode_params(_BAD_ESCAPES_FROM_UNICODE),
    _bad_grammar_params(_BAD_ESCAPES_FROM_BYTES),
    _paired_params(_INCOMPLETE_ESCAPES, 'INCOMPLETE ESCAPES'),
    _good_params(_UNSPACED_SEXPS),
    _paired_params(_SKIP, 'SKIP'),
    _paired_params(_GOOD_FLUSH, 'GOOD FLUSH'),
    _paired_params(_BAD_FLUSH, 'BAD FLUSH'),
    # All top-level values as individual data events, space-delimited.
    _top_level_value_params(),
    # All top-level values as one data event, space-delimited.
    all_top_level_as_one_stream_params(_scalar_iter, (b' ', False)),
    # All top-level values as one data event, block comment-delimited.
    all_top_level_as_one_stream_params(_scalar_iter, (b'/*foo*/', False)),
    # All top-level values as one data event, line comment-delimited.
    all_top_level_as_one_stream_params(_scalar_iter, (b'//foo\n', False)),
    # All annotated top-level values, space-delimited.
    _annotate_params(_top_level_value_params(is_delegate=True)),
    # All annotated top-level values, comment-delimited.
    _annotate_params(_top_level_value_params(b'//foo\n/*bar*/', is_delegate=True)),
    _annotate_params(_good_params(_UNSPACED_SEXPS, is_delegate=True)),
    # All values, each as the only value within a container.
    _containerize_params(_scalar_params()),
    _containerize_params(_containerize_params(_scalar_params(), is_delegate=True, top_level=False), with_skip=False),
    # All values, annotated, each as the only value within a container.
    _containerize_params(_annotate_params(_scalar_params(), is_delegate=True)),
    # All values within a single container.
    _containerize_params(_all_scalars_in_one_container_params()),
    # Annotated containers.
    _containerize_params(_annotate_params(_all_scalars_in_one_container_params(), is_delegate=True)),
    # All unspaced sexps, annotated, in containers.
    _containerize_params(_annotate_params(_incomplete_params(
        _UNSPACED_SEXPS, is_delegate=True, top_level=False), is_delegate=True
    )),
))
def test_raw_reader(p):
    reader_scaffold(reader(is_unicode=p.is_unicode), p.event_pairs)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Provides symbol table managed processing for Ion readers."""
from typing import NamedTuple

from .core import IonEventType, IonType, IonThunkEvent, MemoizingThunk, Transition, \
                  ION_VERSION_MARKER_EVENT
from .exceptions import IonException
from .reader import NEXT_EVENT, SKIP_EVENT
from .symbols import SymbolTable, SymbolTableCatalog, \
                     LOCAL_TABLE_TYPE, SYSTEM_SYMBOL_TABLE, \
                     TEXT_ION, TEXT_ION_1_0, TEXT_ION_SYMBOL_TABLE, TEXT_SYMBOLS, TEXT_IMPORTS, \
                     TEXT_NAME, TEXT_VERSION, TEXT_MAX_ID
from .util import coroutine


class _ManagedContext(NamedTuple):
    """Context for the managed reader.

    Args:
        catalog (SymbolTableCatalog): The catalog for this context.
        symbol_table (SymbolTable): The symbol table.
    """
    catalog: SymbolTableCatalog
    symbol_table: SymbolTable = SYSTEM_SYMBOL_TABLE

    def resolve(self, token):
        """Attempts to resolve the :class:`SymbolToken` against the current table.

        If the ``text`` is not None, the token is returned, otherwise, a token
        in the table is attempted to be retrieved.  If not token is found, then
        this method will raise.
        """
        if token.text is not None:
            return token
        resolved_token = self.symbol_table.get(token.sid, None)
        if resolved_token is None:
            raise IonException('Out of range SID: %d' % token.sid)
        return resolved_token

    def has_symbol_table_annotation(self, annotations):
        if len(annotations) == 0:
            return False

        first = self.resolve(annotations[0])
        return first.text == TEXT_ION_SYMBOL_TABLE


class _IonManagedThunkEvent(IonThunkEvent):
    """An :class:`IonEvent` whose ``value`` field is a thunk."""
    def __new__(cls, event_type, ion_type, value, field_name, annotations, depth):
        return super(_IonManagedThunkEvent, cls).__new__(
            cls, event_type, ion_type,
            value, MemoizingThunk(field_name), MemoizingThunk(annotations),
            depth
        )

    @property
    def field_name(self):
        # We're masking the field_name attribute, this gets around that.
        return self[3]()

    @property
    def annotations(self):
        # We're masking the annotations attribute, this gets around that.
        return self[4]()

    def derive_annotations(self, annotations):
        annotations_thunk = annotations
        if not callable(annotations):
            def annotations_thunk():
                return annotations

        return IonThunkEvent.derive_annotations(self, annotations_thunk)

    def derive_field_name(self, field_name):
        field_name_thunk = field_name
        if not callable(field_name):
            def field_name_thunk():
                return field_name

        return IonThunkEvent.derive_field_name(self, field_name_thunk)


def _managed_thunk_event(ctx, ion_event):
    event_type = ion_event.event_type
    ion_type = ion_event.ion_type

    def field_name_thunk():
        field_name = ion_event.field_name
        if field_name is not None:
            field_name = ctx.resolve(field_name)
        return field_name

    def annotations_thunk():
        return tuple(ctx.resolve(annotation) for annotation in ion_event.annotations)

    def value_thunk():
        value = ion_event.value
        if ion_type is IonType.SYMBOL and value is not None:
            value = ctx.resolve(value)
        return value

    return _IonManagedThunkEvent(
        event_type,
        ion_type,
        value_thunk,
        field_name_thunk,
        annotations_thunk,
        ion_event.depth
    )


@coroutine
def _symbols_handler(symbols, whence):
    ion_event, self = yield
    while True:
        event_type = ion_event.event_type
        ion_type = ion_event.ion_type

        if event_type is IonEventType.CONTAINER_END:
            yield Transition(NEXT_EVENT, whence)

        if event_type is IonEventType.CONTAINER_START:
            # We need to skip past this container (ignoring the end container event).
            ion_event, _ = yield Transition(SKIP_EVENT, self)

        if ion_type is IonType.STRING:
            # Add the potentially null text.
            symbols.append(ion_event.value)

        ion_event, _ = yield Transition(NEXT_EVENT, self)


class _ImportDesc(object):
    __slots__ = [TEXT_NAME, TEXT_VERSION, TEXT_MAX_ID]

    def __init__(self, name=None, version=1, max_id=None):
        self.name = name
        self.version = version
        self.max_id = max_id

    def __str__(self):
        return '_ImportDesc(%s, %s, %s)' % (self.name, self.version, self.max_id)


@coroutine
def _import_desc_handler(ctx, imports, whence):
    ion_event, self = yield

    desc = _ImportDesc()
    while True:
        event_type = ion_event.event_type
        ion_type = ion_event.ion_type

        if event_type is IonEventType.CONTAINER_END:
            if desc.name is not None or desc.name == TEXT_ION:
                table = ctx.catalog.resolve(desc.name, desc.version, desc.max_id)
                imports.append(table)

            yield Transition(NEXT_EVENT, whence)

        if event_type is IonEventType.CONTAINER_START:
            # We need to skip past this container (ignoring the end container event).
            ion_event, _ = yield Transition(SKIP_EVENT, self)
        else:
            field_name = ctx.resolve(ion_event.field_name).text
            if field_name == TEXT_NAME and ion_type is IonType.STRING:
                desc.name = ion_event.value
            elif field_name == TEXT_VERSION and ion_type is IonType.INT:
                desc.version = ion_event.value
            elif field_name == TEXT_MAX_ID and ion_type is IonType.INT:
                desc.max_id = ion_event.value

        ion_event, _ = yield Transition(NEXT_EVENT, self)


@coroutine
def _imports_handler(ctx, imports, whence):
    ion_event, self = yield
    while True:
        event_type = ion_event.event_type
        ion_type = ion_event.ion_type

        if event_type is IonEventType.CONTAINER_END:
            yield Transition(NEXT_EVENT, whence)

        trans = Transition(NEXT_EVENT, self)
        if event_type is IonEventType.CONTAINER_START:
            if ion_type is IonType.STRUCT:
                trans = Transition(NEXT_EVENT, _import_desc_handler(ctx, imports, self))
            else:
                # We need to skip past this container (ignoring the end container event).
                ion_event, _ = yield Transition(SKIP_EVENT, self)

        ion_event, _ = yield trans


@coroutine
def _local_symbol_table_handler(ctx):
    symbols = []
    imports = None

    ion_event, self = yield
    while True:
        if ion_event.event_type is IonEventType.CONTAINER_END:
            break

        event_type = ion_event.event_type
        ion_type = ion_event.ion_type
        if ion_event.field_name is None:
            field_name = None
        else:
            field_name = ctx.resolve(ion_event.field_name).text

        trans = Transition(NEXT_EVENT, self)
        if event_type is IonEventType.CONTAINER_START:
            if field_name == TEXT_SYMBOLS and ion_type is IonType.LIST:
                # Yield to handling local symbol list.
                trans = Transition(NEXT_EVENT, _symbols_handler(symbols, self))

            elif field_name == TEXT_IMPORTS and ion_type is IonType.LIST:
                # Yield to handling imports.
                imports = []
                trans = Transition(NEXT_EVENT, _imports_handler(ctx, imports, self))

            else:
                # We need to skip past this container (ignoring the end container event).
                ion_event, _ = yield Transition(SKIP_EVENT, self)
        elif ion_type == IonType.SYMBOL \
                and field_name == TEXT_IMPORTS \
                and ctx.resolve(ion_event.value).text == TEXT_ION_SYMBOL_TABLE:
            if ctx.symbol_table.table_type.is_system:
                # Force the imports to nothing (system tables import implicitly).
                imports = None
            else:
                # Set the imports to the previous local symbol table.
                imports = [ctx.symbol_table]

        ion_event, _ = yield trans

    # Construct the resulting context and terminate the processing.
    symbol_table = SymbolTable(LOCAL_TABLE_TYPE, symbols, imports=imports)
    yield Transition(_ManagedContext(ctx.catalog, symbol_table), None)


@coroutine
def managed_reader(reader, catalog=None):
    """Managed reader wrapping another reader.

    Args:
        reader (Coroutine): The underlying non-blocking reader co-routine.
        catalog (Optional[SymbolTableCatalog]): The catalog to use for resolving imports.

    Yields:
        Events from the underlying reader delegating to symbol table processing as needed.
        The user will never see things like version markers or local symbol tables.
    """
    if catalog is None:
        catalog = SymbolTableCatalog()

    ctx = _ManagedContext(catalog)
    symbol_trans = Transition(None, None)
    ion_event = None
    while True:
        if symbol_trans.delegate is not None \
                and ion_event is not None \
                and not ion_event.event_type.is_stream_signal:
            # We have a symbol processor active, do not yield to user.
            delegate = symbol_trans.delegate
            symbol_trans = delegate.send(Transition(ion_event, delegate))
            if symbol_trans.delegate is None:
                # When the symbol processor terminates, the event is the context
                # and there is no delegate.
                ctx = symbol_trans.event
                data_event = NEXT_EVENT
            else:
                data_event = symbol_trans.event
        else:
            data_event = None

            if ion_event is not None:
                event_type = ion_event.event_type
                ion_type = ion_event.ion_type
                depth = ion_event.depth

                # System values only happen at the top-level
                if depth == 0:
                    if event_type is IonEventType.VERSION_MARKER:
                        if ion_event != ION_VERSION_MARKER_EVENT:
                            raise IonException('Invalid IVM: %s' % (ion_event,))

                        # Reset and swallow IVM
                        ctx = _ManagedContext(ctx.catalog)
                        data_event = NEXT_EVENT

                    elif ion_type is IonType.SYMBOL \
                            and len(ion_event.annotations) == 0 \
                            and ion_event.value is not None \
                            and ctx.resolve(ion_event.value).text == TEXT_ION_1_0:
                        assert symbol_trans.delegate is None

                        # A faux IVM is a NOP
                        data_event = NEXT_EVENT

                    elif event_type is IonEventType.CONTAINER_START \
                            and ion_type is IonType.STRUCT \
                            and ctx.has_symbol_table_annotation(ion_event.annotations):
                        assert symbol_trans.delegate is None

                        # Activate a new symbol processor.
                        delegate = _local_symbol_table_handler(ctx)
                        symbol_trans = Transition(None, delegate)
                        data_event = NEXT_EVENT

            if data_event is None:
                # No system processing or we have to get data, yield control.
                if ion_event is not None:
                    ion_event = _managed_thunk_event(ctx, ion_event)
                data_event = yield ion_event

        ion_event = reader.send(data_event)





# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Tests the `equivalence` module."""

from decimal import Decimal
from functools import partial
from itertools import chain

from datetime import datetime, timedelta

from amazon.ion.core import IonType, timestamp, TimestampPrecision, OffsetTZInfo
from amazon.ion.equivalence import ion_equals, obj_has_ion_type_and_annotation
from amazon.ion.simple_types import IonPyNull, IonPyInt, IonPyBool, IonPyFloat, IonPyDecimal, IonPyText, IonPyBytes, \
    IonPyList, IonPyTimestamp, IonPySymbol, IonPyDict
from amazon.ion.symbols import SymbolToken, ImportLocation
from tests import parametrize


def _ion_nature(_ion_nature_cls, ion_type, value=None, annotations=()):
    return _ion_nature_cls.from_value(ion_type, value, annotations)


_IT = IonType
_D = Decimal
_dt = datetime
_ts = timestamp
_TP = TimestampPrecision
_ST = SymbolToken

_null = partial(_ion_nature, IonPyNull)
_bool = partial(_ion_nature, IonPyInt, _IT.BOOL)
_int = partial(_ion_nature, IonPyBool, _IT.INT)
_float = partial(_ion_nature, IonPyFloat, _IT.FLOAT)
_decimal = partial(_ion_nature, IonPyDecimal, _IT.DECIMAL)
_timestamp = partial(_ion_nature, IonPyTimestamp, _IT.TIMESTAMP)
_string = partial(_ion_nature, IonPyText, _IT.STRING)
_symbol_text = partial(_ion_nature, IonPyText, _IT.SYMBOL)
_symbol_token = partial(_ion_nature, IonPySymbol, _IT.SYMBOL)
_clob = partial(_ion_nature, IonPyBytes, _IT.CLOB)
_blob = partial(_ion_nature, IonPyBytes, _IT.BLOB)
_list = partial(_ion_nature, IonPyList, _IT.LIST)
_sexp = partial(_ion_nature, IonPyList, _IT.SEXP)
_struct = partial(_ion_nature, IonPyDict, _IT.STRUCT)


_EQUIVS = (
    # For each tuple, all elements in that tuple are equivalent under the Ion data model.
    (None, _null(_IT.NULL), _null(_IT.NULL)),
    (True, _bool(True), _bool(1)),
    (False, _bool(0), _bool(False)),
    (0, _int(0), _int(0)),
    (0, -0, _int(0), _int(-0)),
    (1, _int(1), _int(1)),
    (-1, _int(-1), _int(-1)),
    (0.0, 0e0, 0., 0., 0e1, 0e-1),
    (1.0, 1.0),
    (float('nan'), float('nan')),
    (float('+inf'), float('+inf')),
    (float('-inf'), float('-inf')),
    (_D(0), _D('0'), _decimal(_D(0)), _decimal(_D('0')), _D('0.'), _decimal(_D('0.'))),
    (_D('0.0'), _decimal(_D('0.0')), _D('0e-1'), _decimal(_D('0e-1'))),
    (_D('1.'), _decimal(_D('1e0')), _D('1e-0')),
    (_D('1e2'), _decimal(_D('1e2'))),
    (_D('-0.'), _D('-0'), _decimal(_D('-0.')), _decimal(_D('-0'))),
    (
        # Regular datetimes are always seconds precision with 6 digits of fractional precision.
        _dt(1, 1, 1),
        _dt(1, 1, 1, 0, 0, 0),
        _timestamp(_dt(1, 1, 1, 0, 0, 0, 0, tzinfo=None)),
        _ts(1, microsecond=0, precision=_TP.SECOND, fractional_precision=6),
        _timestamp(_ts(1, microsecond=0, precision=_TP.SECOND)),
        _ts(1, precision=_TP.SECOND, fractional_seconds=Decimal('0.000000')),
    ),
    (
        u'abc',
        _string(u'abc')
    ),
    (
        # Having the same text makes all these equivalent, regardless of symbol ID or import location.
        u'abc',
        _symbol_text(u'abc'),
        _symbol_token(_ST(u'abc', None)),
        _symbol_token(_ST(u'abc', 10)),
        _ST(u'abc', 10),
        _ST(u'abc', None),
        _ST(u'abc', 10, ImportLocation(u'foo', 1)),
        _symbol_token(_ST(u'abc', 10, ImportLocation(u'bar', 2)))
    ),
    (
        # Symbols with unknown text from a local symbol table context are equivalent to each other.
        _symbol_token(_ST(None, 10)),
        _symbol_token(_ST(None, 11)),
        _ST(None, 12),
        _ST(None, 13),
    ),
    (
        _symbol_token(_ST(None, 0)),
        _ST(None, 0),
    ),
    (
        b'abc',
        _blob(b'abc'),
    ),
    (
        b'abc',
        _clob(b'abc'),
    ),
    (
        [],
        _list([]),
    ),
    (
        [],
        _sexp([]),
    ),
    (
        {},
        _struct({}),
    ),
    (
        {u'abc': 123, 'def': 456},
        _struct({u'def': 456, u'abc': 123}),
    )
)

_EQUIVS_INSTANTS = (
    (
        # These all represent the same instant.
        _dt(2000, 1, 1),
        _dt(2000, 1, 1, tzinfo=(OffsetTZInfo())),
        _dt(2000, 1, 1, 1, tzinfo=OffsetTZInfo(timedelta(hours=1))),
        _ts(1999, 12, 31, 23, 59, off_hours=0, off_minutes=-1, precision=_TP.SECOND),
        _timestamp(_ts(2000, 1, 1, 1, off_hours=1, off_minutes=0, precision=_TP.SECOND)),
        _timestamp(_ts(2000, 1, 1, 1, off_hours=1, off_minutes=0, precision=_TP.SECOND,
                       fractional_seconds=Decimal('0.000000'))),
    ),
)

_NONEQUIVS = (
    # For each tuple, each element is not equivalent to any other element
    # under the Ion data model.
    (None, 0, _null(_IT.BOOL)),
    (True, False),
    (True, _bool(False)),
    (_bool(True), False),
    (_bool(True), _bool(False)),
    (_bool(True), _int(1)),  # True == 1 in python but not in the ion datamodel
    (1, -1, 1.),
    (1, _int(-1), _float(1)),
    (_int(1), _int(-1)),
    (1, _int(-1)),
    (1., 1),
    (-0., 0.),
    (_float(-0.), _float(0.), _int(0)),
    (-0., _float(0.), 0, None),
    (_float(0e1), _float(-0e1)),
    (float('+inf'), float('-inf')),
    (_float(float('+inf')), _float(float('-inf'))),
    (float('+inf'), _float(float('-inf'))),
    (_float(float('nan')), _float(float('+inf'))),
    (float('nan'), _float(float('+inf')), None),
    (_D('-0.'), _D(0), None),
    (_D('-0'), _decimal(_D(0))),
    (_decimal(_D('-0')), _decimal(_D('0.'))),
    (_D('0.'), _D('0e1'), _decimal(_D('0e-1'))),
    (_D('1.'), _D('1.0'), _decimal(_D('1.00'))),
    (
        # Timestamps with different precisions are not equivalent.
        _dt(1, 1, 1),
        _timestamp(_ts(1, 1, 1, precision=_TP.DAY)),
        _ts(1, microsecond=0, precision=_TP.SECOND, fractional_precision=3),
        _dt(1, 1, 1, tzinfo=OffsetTZInfo()),
        _timestamp(_ts(1, microsecond=0, precision=_TP.SECOND, fractional_precision=6, off_hours=-1, off_minutes=0)),
        None
    ),
    (
        # Timestamps that represent the same instant with different offsets are not equivalent.
        _dt(2000, 1, 1),
        _dt(2000, 1, 1, tzinfo=OffsetTZInfo()),
        _timestamp(_ts(2000, 1, 1, 1, off_hours=1, off_minutes=0, precision=_TP.SECOND)),
        _ts(1999, 12, 31, 23, 59, off_hours=0, off_minutes=-1, precision=_TP.SECOND),
        None,
    ),
    (
        # Timestamps with different fractional seconds are not equivalent.
        _timestamp(_ts(2000, 1, 1, 1, precision=_TP.SECOND, fractional_seconds=Decimal('0.123456789'))),
        _timestamp(_ts(2000, 1, 1, 1, precision=_TP.SECOND, fractional_seconds=Decimal('0.9999999'))),
        _timestamp(_ts(2000, 1, 1, 1, precision=_TP.SECOND, fractional_seconds=Decimal('0.000000000'))),
        None,
    ),
    (
        _string(u'abc'),
        u'abcd',
        _symbol_text(u'abc'),
        _symbol_text(u'abcde'),
        _symbol_token(_ST(None, 10)),
        _symbol_token(_ST(u'abcdef', 10)),
        _ST(u'abcdefg', 10),
        None,
    ),
    (
        _ST(None, 10, ImportLocation(u'foo', 1)),
        _ST(None, 10, ImportLocation(u'foo', 2)),
        _symbol_token(_ST(None, 10, ImportLocation(u'bar', 1))),
        _symbol_token(_ST(None, 10, ImportLocation(u'bar', 2))),
        _symbol_token(_ST(None, 10)),
        _ST(None, 0),  # Symbol 0 is only equivalent to itself.
        None,
    ),
    (
        _blob(b'abc'),
        _clob(b'abc'),
        None,
    ),
    (
        _list([]),
        _sexp([]),
        {},
        None
    ),
    (
        _list([]),
        [u'abc'],
        [u'def', u'abc'],
        _list([u'abc', u'def'])
    ),
    (
        _sexp([]),
        [u'abc'],
        [u'def', u'abc'],
        _sexp([u'abc', u'def'])
    ),
    (
        _struct({}),
        {u'abc': 123}
    ),
    (
        {u'abc': 456, u'def': 123},
        {u'abc': 123, u'def': 456},
        _struct({u'abc': 123, u'def': 123}),
        _struct({u'abc': 456, u'def': 456}),
    )
)


class _Parameter:
    def __init__(self, desc, assertion):
        self.desc = desc
        self.assertion = assertion

    def __str__(self):
        return self.desc

_P = _Parameter


def _desc(a, b, operator):
    def _str(val):
        if obj_has_ion_type_and_annotation(val):
            return '%s(%s, ion_type=%s, ion_annotations=%s)' % (type(val), val, val.ion_type, val.ion_annotations)
        return '%s(%s)' % (type(val), val)
    return 'assert %s %s %s' % (_str(a), operator, _str(b))


def _equivs_param(a, b, timestamp_instants_only=False):
    def assert_equivalent():
        assert ion_equals(a, b, timestamp_instants_only)
        assert ion_equals(b, a, timestamp_instants_only)
    return _P(_desc(a, b, '=='), assert_equivalent)


_equivs_timestamp_instants_param = partial(_equivs_param, timestamp_instants_only=True)


def _nonequivs_param(a, b):
    def assert_not_equivalent():
        assert not ion_equals(a, b)
        assert not ion_equals(b, a)
    return _P(_desc(a, b, '!='), assert_not_equivalent)


_TEST_ANNOTATIONS = (
    (u'abc',),
    (u'abc', u'def'),
    (SymbolToken(text=None, sid=10), SymbolToken(text=None, sid=11)),
)


def _generate_annotations():
    """Circularly generates sequences of test annotations. The annotations sequence yielded from this generator must
    never be equivalent to the annotations sequence last yielded by this generator.
    """
    i = 0
    while True:
        yield _TEST_ANNOTATIONS[i]
        i += 1
        if i == len(_TEST_ANNOTATIONS):
            i = 0

_annotations_generator = iter(_generate_annotations())


def _add_annotations(val):
    val.ion_annotations = next(_annotations_generator)


def _generate_equivs(equivs, param_func=_equivs_param):
    for seq in equivs:
        seq_len = len(seq)
        for i in range(seq_len):
            for j in range(seq_len):
                yield param_func(seq[i], seq[j])
                # Now, for pairs that are otherwise equivalent, add differing annotation(s) if possible and make sure
                # they're no longer equivalent.
                has_ion_nature = False
                a = seq[i]
                b = seq[j]
                if obj_has_ion_type_and_annotation(a):
                    has_ion_nature = True
                    a = a.__copy__()
                    _add_annotations(a)
                if obj_has_ion_type_and_annotation(b):
                    has_ion_nature = True
                    b = b.__copy__()
                    _add_annotations(b)
                if has_ion_nature:
                    yield _nonequivs_param(a, b)


def _generate_nonequivs(nonequivs):
    for seq in nonequivs:
        seq_len = len(seq)
        for i in range(seq_len):
            for j in range(seq_len):
                if i != j:
                    yield _nonequivs_param(seq[i], seq[j])


def list_from(equivs, equiv_set_index):
    output = []
    for equiv_set in equivs:
        output.append(equiv_set[equiv_set_index])
    return output


def _generate_equiv_lists(equivs, type_func):
    a = list_from(equivs, 0)
    b = list_from(equivs, 1)
    yield _equivs_param(a, b)
    yield _equivs_param(a, type_func(b))
    yield _equivs_param(type_func(a), type_func(b))


def _generate_equiv_dicts(equivs):
    a = list_from(equivs, 0)
    b = list_from(equivs, 1)
    assert len(a) == len(b)
    field_names = [u'%d' % (i,) for i in range(len(a))]
    a = dict(zip(field_names, a))
    b = dict(zip(field_names, b))
    yield _equivs_param(a, b)
    yield _equivs_param(a, _struct(b))
    yield _equivs_param(_struct(a), _struct(b))


@parametrize(
    *tuple(chain(
        _generate_equivs(_EQUIVS),
        _generate_equivs(_EQUIVS_INSTANTS, _equivs_timestamp_instants_param),
        _generate_nonequivs(_NONEQUIVS),
        _generate_equiv_lists(_EQUIVS, _list),
        _generate_equiv_lists(_EQUIVS, _sexp),
        _generate_equiv_dicts(_EQUIVS),
    ))
)
def test_equivalence(p):
    p.assertion()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from typing import Callable, Sequence, NamedTuple

from pytest import raises

from tests import parametrize

from amazon.ion.reader import BufferQueue, CodePointArray


def read(expected):
    def action(queue):
        if isinstance(expected, int):
            read_len = 1
            actual = queue.read_byte()
        else:
            read_len = len(expected)
            actual = queue.read(read_len)
        assert actual == expected

        return -read_len, read_len

    return action


def read_byte(ch):
    return read(ord(ch))


def unread(data):
    def action(queue):
        queue.unread(data)
        if isinstance(data, int):
            unread_len = 1
        else:
            unread_len = len(data)
        return unread_len, -unread_len

    return action


def unread_byte(ch):
    return unread(ord(ch))


def skip(amount, expected_rem=0):
    def action(queue):
        rem = queue.skip(amount)
        skip_amount = amount - rem
        assert expected_rem == rem
        return -skip_amount, skip_amount

    return action


def extend(data):
    def action(queue):
        queue.extend(data)

        return len(data), 0

    return action


def expect(exception, action):
    def raises_action(queue):
        with raises(exception):
            action(queue)
        return 0, 0

    return raises_action


def mark_eof():
    def action(queue):
        queue.mark_eof()
        return 1, 0

    return action


def expect_eof(is_eof):
    def action(queue):
        maybe_eof = queue.read_byte()
        assert is_eof is BufferQueue.is_eof(maybe_eof)
        return -1, 1

    return action


class _P(NamedTuple):
    desc: str
    actions: Sequence[Callable]
    is_unicode: bool = False

    def __str__(self):
        return self.desc


@parametrize(
    _P(
        desc='EMPTY READ BYTE',
        actions=[
            expect(IndexError, read_byte(b'a')),
        ],
    ),
    _P(
        desc='EMPTY READ',
        actions=[
            expect(IndexError, read(b'a')),
        ],
    ),
    _P(
        desc='EMPTY SKIP',
        actions=[
            skip(1, expected_rem=1),
        ],
    ),
    _P(
        desc='EMPTY UNREAD',
        actions=[
            expect(IndexError, unread_byte(b'a')),
        ],
    ),
    _P(
        desc='INCORRECT UNREAD',
        actions=[
            extend(b'ab'),
            read_byte(b'a'),
            expect(ValueError, unread_byte(b'c')),
        ],
    ),
    _P(
        desc='SINGLE FULL',
        actions=[
            extend(b'abcd'),
            read(b'abcd'),
        ],
    ),
    _P(
        desc='SINGLE BYTE FULL',
        actions=[
            extend(b'a'),
            read_byte(b'a'),
        ],
    ),
    _P(
        desc='SINGLE SKIP FULL',
        actions=[
            extend(b'abcd'),
            skip(4, expected_rem=0),
            skip(4, expected_rem=4),
        ],
    ),
    _P(
        desc='SINGLE PART',
        actions=[
            extend(b'abcdefg'),
            read(b'ab'),
            read_byte(b'c'),
            skip(2),
            read(b'fg'),
        ],
    ),
    _P(
        desc='SINGLE UNREAD',
        actions=[
            extend(b'a'),
            read_byte(b'a'),
            unread_byte(b'a'),
            read_byte(b'a')
        ]
    ),
    _P(
        desc='MULTI CODE UNIT UNREAD',
        actions=[
            extend(u'a\U0001f4a9c'),
            read(u'a\U0001f4a9'),
            unread(u'\U0001f4a9'),
            read(u'\U0001f4a9c')
        ],
        is_unicode=True
    ),
    _P(
        desc='MULTI ALL',
        actions=[
            extend(b'abcd'),
            extend(b'efgh'),
            read(b'abcdefgh'),

            extend(b'ijkl'),
            extend(b'mnop'),
            read(b'ijklmnop'),

            extend(b'ab'),
            extend(b'cd'),
            read(b'abc'),
            unread_byte(b'c'),
            unread_byte(b'b'),
            read(b'bcd'),
            extend(b'EF'),
            extend(b'GH'),
            extend(b'ef'),
            extend(b'gh'),
            skip(4),
            read_byte(b'e'),
            read_byte(b'f'),
            read_byte(b'g'),
            read_byte(b'h'),
            unread_byte(b'h'),
            read_byte(b'h')
        ],
    ),
    _P(
        desc='MULTI ALL UNICODE',
        actions=[
            extend(u'abcd'),
            extend(u'efgh'),
            read(u'abcdefgh'),

            extend(u'ijkl'),
            extend(u'mnop'),
            read(u'ijklmnop'),

            extend(u'ab'),
            extend(u'cd'),
            read(u'abc'),
            unread_byte(u'c'),
            unread_byte(u'b'),
            read(u'bcd'),
            extend(u'EF'),
            extend(u'GH'),
            extend(u'ef'),
            extend(u'gh'),
            skip(4),
            read_byte(u'e'),
            read_byte(u'f'),
            read_byte(u'g'),
            read_byte(u'h'),
            unread_byte(u'h'),
            read_byte(u'h')
        ],
        is_unicode=True
    ),
    _P(
        desc='MULTI PART AND SPAN',
        actions=[
            extend(b'abcd'),
            extend(b'efgh'),
            read(b'ab'),
            read(b'cdef'),

            extend(b'ijkl'),
            read(b'ghij'),

            extend(b'mnop'),
            read(b'klmnop'),

            extend(b'abcd'),
            extend(b'efgh'),
            extend(b'ijkl'),
            extend(b'm'),
            read(b'abcdefghi'),
            read_byte(b'j'),
            read_byte(b'k'),
            read_byte(b'l'),
            read_byte(b'm'),
        ],
    ),
    _P(
        desc='MULTI MIDDLE',
        actions=[
            extend(b'abcdefg'),
            extend(b'hijklmn'),
            read(b'ab'),
            read(b'cdef'),
            read(b'ghij'),
            read(b'klm'),
            read(b'n')
        ],
    ),
    _P(
        desc='MULTI MIDDLE SKIP',
        actions=[
            extend(b'abcdefg'),
            extend(b'hijklmn'),
            read(b'ab'),
            skip(4),
            read(b'ghij'),
            skip(3),
            read(b'n')
        ],
    ),
    _P(
        desc='EOF',
        actions=[
            extend(b'ab'),
            mark_eof(),
            read(b'ab'),
            expect_eof(True)
        ]
    ),
    _P(
        desc='FAKE EOF',
        actions=[
            extend(b'ab\x04'),
            read(b'ab'),
            expect_eof(False)
        ]
    ),
    _P(
        desc='EOF UNICODE',
        actions=[
            extend(u'ab'),
            mark_eof(),
            read(u'ab'),
            expect_eof(True)
        ],
        is_unicode=True
    ),
    _P(
        desc='FAKE EOF UNICODE',
        actions=[
            extend(u'ab\x04'),
            read(u'ab'),
            expect_eof(False)
        ],
        is_unicode=True
    ),
    _P(
        desc='BYTES BUFFER GIVEN UNICODE',
        actions=[
            extend(b'abc'),
            expect(ValueError, extend(u'def'))
        ],
        is_unicode=False
    ),
    _P(
        desc='UNICODE BUFFER GIVEN BYTES',
        actions=[
            extend(u'abc'),
            expect(ValueError, extend(b'def'))
        ],
        is_unicode=True
    ),
    _P(
        desc='BYTES BUFFER UNREAD UNICODE',
        actions=[
            extend(b'a'),
            read_byte(b'a'),
            expect(ValueError, unread(u'a'))
        ],
        is_unicode=False
    ),
)
def test_buffer_queue(p):
    queue = BufferQueue(p.is_unicode)
    expected_len = 0
    expected_pos = 0

    for action in p.actions:
        queue_len = len(queue)
        queue_pos = queue.position
        assert queue_len == expected_len
        assert queue_pos == expected_pos

        len_change, pos_change = action(queue)
        expected_len += len_change
        expected_pos += pos_change



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: amazon/ionbenchmark/self_describing_proto.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/amazon/ionbenchmark/self_describing_proto.proto\x1a\x19google/protobuf/any.proto\x1a google/protobuf/descriptor.proto\"z\n\x15SelfDescribingMessage\x12:\n\x0e\x64\x65scriptor_set\x18\x01 \x01(\x0b\x32\".google.protobuf.FileDescriptorSet\x12%\n\x07message\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Anyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'amazon.ionbenchmark.self_describing_proto_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_SELFDESCRIBINGMESSAGE']._serialized_start=112
  _globals['_SELFDESCRIBINGMESSAGE']._serialized_end=234
# @@protoc_insertion_point(module_scope)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Provides utilities for determining whether two objects are equivalent under the Ion data model."""
import struct
from datetime import datetime
from decimal import Decimal
from math import isnan

from amazon.ion.core import IonType, Timestamp, TimestampPrecision, MICROSECOND_PRECISION, OffsetTZInfo, Multimap
from amazon.ion.simple_types import IonPyList, IonPyDict, IonPyTimestamp, IonPyNull, IonPySymbol, \
    IonPyText, IonPyDecimal, IonPyFloat
from amazon.ion.symbols import SymbolToken


def obj_has_ion_type_and_annotation(obj):
    return hasattr(obj, 'ion_type') and hasattr(obj, 'ion_annotations')


def ion_equals(a, b, timestamps_instants_only=False):
    """Tests two objects for equivalence under the Ion data model.

    There are three important cases:
        * When neither operand specifies its `ion_type` or `annotations`, this method will only return True when the
          values of both operands are equivalent under the Ion data model.
        * When only one of the operands specifies its `ion_type` and `annotations`, this method will only return True
          when that operand has no annotations and has a value equivalent to the other operand under the Ion data model.
        * When both operands specify `ion_type` and `annotations`, this method will only return True when the ion_type
          and annotations of both are the same and their values are equivalent under the Ion data model.

    Note that the order of the operands does not matter.

    Args:
        a (object): The first operand.
        b (object): The second operand.
        timestamps_instants_only (Optional[bool]): False if timestamp objects (datetime and its subclasses) should be
            compared according to the Ion data model (where the instant, precision, and offset must be equal); True
            if these objects should be considered equivalent if they simply represent the same instant.
    """
    if timestamps_instants_only:
        return _ion_equals_timestamps_instants(a, b)
    return _ion_equals_timestamps_data_model(a, b)


def _ion_equals_timestamps_instants(a, b):
    return _ion_equals(a, b, _timestamp_instants_eq, _ion_equals_timestamps_instants)


def _ion_equals_timestamps_data_model(a, b):
    return _ion_equals(a, b, _timestamps_eq, _ion_equals_timestamps_data_model)


def _ion_equals(a, b, timestamp_comparison_func, recursive_comparison_func):
    """Compares a and b according to the description of the ion_equals method."""
    for a, b in ((a, b), (b, a)):  # Ensures that operand order does not matter.
        if obj_has_ion_type_and_annotation(a):
            if obj_has_ion_type_and_annotation(b):
                # Both operands have _IonNature. Their IonTypes and annotations must be equivalent.
                eq = a.ion_type is b.ion_type and _annotations_eq(a, b)
            else:
                # Only one operand has _IonNature. It cannot be equivalent to the other operand if it has annotations.
                eq = not a.ion_annotations
            if eq:
                if isinstance(a, IonPyList):
                    return _sequences_eq(a, b, recursive_comparison_func)
                elif isinstance(a, IonPyDict):
                    return _structs_eq(a, b, recursive_comparison_func)
                elif isinstance(a, IonPyTimestamp):
                    return timestamp_comparison_func(a, b)
                elif isinstance(a, IonPyNull):
                    return isinstance(b, IonPyNull) or (b is None and a.ion_type is IonType.NULL)
                elif isinstance(a, IonPySymbol) or (isinstance(a, IonPyText) and a.ion_type is IonType.SYMBOL):
                    return _symbols_eq(a, b)
                elif isinstance(a, IonPyDecimal):
                    return _decimals_eq(a, b)
                elif isinstance(a, IonPyFloat):
                    return _floats_eq(a, b)
                else:
                    return a == b
            return False
    # Reaching this point means that neither operand has _IonNature.
    for a, b in ((a, b), (b, a)):  # Ensures that operand order does not matter.
        if isinstance(a, list):
            return _sequences_eq(a, b, recursive_comparison_func)
        elif isinstance(a, dict):
            return _structs_eq(a, b, recursive_comparison_func)
        elif isinstance(a, datetime):
            return timestamp_comparison_func(a, b)
        elif isinstance(a, SymbolToken):
            return _symbols_eq(a, b)
        elif isinstance(a, Decimal):
            return _decimals_eq(a, b)
        elif isinstance(a, float):
            return _floats_eq(a, b)
    return a == b


def _annotations_eq(a, b):
    return _sequences_eq(a.ion_annotations, b.ion_annotations, _symbols_eq)


def _sequences_eq(a, b, comparison_func):
    assert isinstance(a, (list, tuple))
    if not isinstance(b, (list, tuple)):
        return False
    sequence_len = len(a)
    if sequence_len != len(b):
        return False
    for i in range(sequence_len):
        if not comparison_func(a[i], b[i]):
            return False
    return True


def _structs_eq(a, b, comparison_func):
    assert isinstance(a, (dict, Multimap, IonPyDict))
    if not isinstance(b, (dict, Multimap, IonPyDict)):
        return False
    dict_len = len(a)
    if dict_len != len(b):
        return False
    for a, b in ((a, b), (b, a)):
        key_iter = iter(a.keys())
        while True:
            try:
                key = next(key_iter)
            except StopIteration:
                break
            if key not in b:
                return False
            if isinstance(a, (IonPyDict, Multimap)) and isinstance(b, (IonPyDict, Multimap)):
                values_a = a.get_all_values(key)
                values_b = b.get_all_values(key)
                if len(values_a) != len(values_b):
                    return False
                for value_a in values_a:
                    if not any(comparison_func(value_a, value_b) for value_b in values_b):
                        return False
            else:
                if not comparison_func(a[key], b[key]):
                    return False

    return True


def _timestamps_eq(a, b):
    """Compares two timestamp operands for equivalence under the Ion data model."""
    assert isinstance(a, datetime)
    if not isinstance(b, datetime):
        return False
    # Local offsets must be equivalent.
    if (a.tzinfo is None) ^ (b.tzinfo is None):
        return False
    if a.utcoffset() != b.utcoffset():
        return False
    for a, b in ((a, b), (b, a)):
        if isinstance(a, Timestamp):
            if isinstance(b, Timestamp):
                # Both operands declare their precisions. They are only equivalent if their precisions are the same.
                if a.precision is b.precision and a.fractional_precision is b.fractional_precision \
                        and a.fractional_seconds == b.fractional_seconds:
                    break
                return False
            elif a.precision is not TimestampPrecision.SECOND or a.fractional_precision != MICROSECOND_PRECISION:
                # Only one of the operands declares its precision. It is only equivalent to the other (a naive datetime)
                # if it has full microseconds precision.
                return False
    return a == b


def _timestamp_instants_eq(a, b):
    """Compares two timestamp operands for point-in-time equivalence only."""
    assert isinstance(a, datetime)
    if not isinstance(b, datetime):
        return False
    # datetime's __eq__ can't compare a None offset and a non-None offset. For these equivalence semantics, a None
    # offset (unknown local offset) is treated equivalently to a +00:00.
    if a.tzinfo is None:
        a = a.replace(tzinfo=OffsetTZInfo())
    if b.tzinfo is None:
        b = b.replace(tzinfo=OffsetTZInfo())
    # datetime's __eq__ implementation compares instants; offsets and precision need not be equal.
    return a == b


def _symbols_eq(a, b):
    assert isinstance(a, (str, SymbolToken))
    if not isinstance(b, (str, SymbolToken)):
        return False
    a_text = getattr(a, 'text', a)
    b_text = getattr(b, 'text', b)
    if a_text == b_text:
        if a_text is None:
            # Both have unknown text. If they come from a local context, they are equivalent.
            a_location = getattr(a, 'location', None)
            b_location = getattr(b, 'location', None)
            if (a_location is None) ^ (b_location is None):
                return False
            if a_location is not None:
                # Both were imported from shared symbol tables. In this case, they are only equivalent if they were
                # imported from the same position in the same shared symbol table.
                if (a_location.name != b_location.name) or (a_location.position != b_location.position):
                    return False
            a_sid = getattr(a, 'sid', None)
            b_sid = getattr(b, 'sid', None)
            if a_sid is None or b_sid is None:
                raise ValueError('Attempted to compare malformed symbols %s, %s.' % (a, b))
            if (a_sid == 0) ^ (b_sid == 0):
                # SID 0 is only equal to SID 0.
                return False
        return True
    return False


def _decimals_eq(a, b):
    assert isinstance(a, Decimal)
    if not isinstance(b, Decimal):
        return False
    if a.is_zero() and b.is_zero():
        if a.is_signed() ^ b.is_signed():
            # Negative-zero is not equivalent to positive-zero.
            return False
    # This ensures that both have equal precision.
    return a.canonical().compare_total(b.canonical()) == 0


def _is_float_negative_zero(x):
    return struct.pack('>d', x) == b'\x80\x00\x00\x00\x00\x00\x00\x00'


def _floats_eq(a, b):
    assert isinstance(a, float)
    if not isinstance(b, float):
        return False
    if a == 0 and b == 0:
        # Negative-zero is not equivalent to positive-zero.
        return not (_is_float_negative_zero(a) ^ _is_float_negative_zero(b))
    # nan is always equivalent to nan.
    return a == b or (isnan(a) and isnan(b))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


class ProtoSerde:
    """
    This class provides load(s) and dump(s) functions that wrap the protocol buffer APIs for a single message type.

    Example Usage:
    ```
    car_proto = proto.ProtoSerde(proto.get_message_type_from_py("Car", "generated_model.car_pb2"))
    truck_proto = proto.ProtoSerde(proto.get_message_type_from_descriptor_set("Truck", "~/trucks_descriptors.desc"))

    car = car_proto.load("~/my_car.data")
    truck = convert_to_truck(car)
    truck_proto.dump(truck, "~/my_truck.data")
    ```
    """
    def __init__(self, message_type, reuse_message_object=True):
        """
        :param message_type: A protocol buffer message class definition that extends the abstract base class Message
            (https://googleapis.dev/python/protobuf/latest/google/protobuf/message.html#google.protobuf.message.Message).
            A Message implementation can be obtained using `protoc`. To load a message at runtime, see
            `get_message_type_from_descriptor_set` and `get_message_type_from_py`
        :param reuse_message_object: Controls whether to reuse the message object. Python protocol buffer message
            objects are designed to be reused to reduce allocations, but you may not want to reuse the message object if
            you need to hold multiple messages in memory or if you are sharing the message objects or an instance of
            this class across multiple threads.
        """
        self._message_type = message_type
        if reuse_message_object:
            self._message_obj = message_type()

            def get_message_instance():
                return self._message_obj

            self._get_message_instance = get_message_instance
        else:
            self._get_message_instance = message_type

    def loads(self, s):
        """
        Deserialize a protocol buffer message from bytes/string.
        """
        obj = self._get_message_instance()
        obj.ParseFromString(s)
        return obj

    def load(self, fp):
        """
        Deserialize a protocol buffer message from a file.
        """
        return self.loads(fp.read())

    def dumps(self, obj):
        """
        Serialize a protocol buffer message to bytes/string.
        """
        return obj.SerializeToString()

    def dump(self, obj, fp):
        """
        Serialize a protocol buffer message to a file.
        """
        fp.write(obj.SerializeToString())


def get_message_type_from_py(type_name, module_name, module_file=None):
    """
    Import and load a protocol buffer message class from a Python module at runtime.

    A Python protocol buffer module can be created using `protoc --python_out=...`.

    :param type_name: Name of the generated Protocol Buffer Message class to get
    :type type_name: str
    :param module_name: Absolute module path of the python module containing the message class
    :type module_name: str
    :param module_file: Location of the module in the file system. Only required if not in sys.path
    :type module_file: str
    :return: a message type to use for ProtoSerde
    """
    import importlib
    import importlib.util
    import sys

    if module_file:
        spec = importlib.util.spec_from_file_location(module_name, module_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    elif module_name:
        module = importlib.import_module(module_name)
    else:
        raise ValueError("Both module_name and module_file are None; at least one must be set.")

    return getattr(module, type_name)


def get_message_type_from_descriptor_set(type_name, descriptor_set_file):
    """
    Dynamically generate a protocol buffer message class definition from a protocol buffer descriptor file.

    A descriptor file can be created using `protoc --descriptor_set_out=...`.

    :param type_name: Name of the type to use from the descriptor set
    :type type_name: str
    :param descriptor_set_file: The path of the protocol buffer descriptor set file
    :type descriptor_set_file: str
    :return: a message type to use for ProtoSerde
    """
    from google.protobuf import descriptor_pb2 as _descriptor_pb2
    from google.protobuf import message as _message
    from google.protobuf import message_factory as _message_factory
    from google.protobuf import reflection as _reflection
    import six

    descriptor_set = _descriptor_pb2.FileDescriptorSet()

    with open(descriptor_set_file, "rb") as f:
        descriptor_set.ParseFromString(f.read())

    messages_types = _message_factory.GetMessages(descriptor_set.file)
    message_type = messages_types[type_name]()

    class DynamicMessage(six.with_metaclass(_reflection.GeneratedProtocolMessageType, _message.Message)):
        DESCRIPTOR = message_type.DESCRIPTOR

    return DynamicMessage



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from itertools import chain
from typing import Sequence, Any, NamedTuple

from tests import parametrize, listify
from tests.reader_util import reader_scaffold, add_depths
from tests.event_aliases import *

from amazon.ion.exceptions import IonException, CannotSubstituteTable
from amazon.ion.reader_managed import managed_reader, _ImportDesc, _IonManagedThunkEvent
from amazon.ion.symbols import shared_symbol_table, local_symbol_table, \
                               SymbolToken, ImportLocation, \
                               SymbolTableCatalog, \
                               SYSTEM_SYMBOL_TABLE, \
                               TEXT_ION, TEXT_ION_1_0, TEXT_ION_SYMBOL_TABLE, \
                               TEXT_NAME, TEXT_VERSION, TEXT_MAX_ID, \
                               TEXT_IMPORTS, TEXT_SYMBOLS
from amazon.ion.util import coroutine

_DATA = e_read(b'DUMMY')

_SYS = SYSTEM_SYMBOL_TABLE

_sid = partial(SymbolToken, None)
_tok = SymbolToken
_loc = ImportLocation
_desc = _ImportDesc


def _nextify(iter_func):
    """Creates a decorator that generates event pairs of ``(NEXT_EVENT, event)``
    where ``event`` are instances returned by the decorated function.
    """
    def delegate(*args, **kw_args):
        for ion_event in iter_func(*args, **kw_args):
            yield NEXT_EVENT, ion_event
    delegate.__name__ = iter_func.__name__
    return delegate


_APPEND = object()


def _system_sid_token(text):
    return _sid(SYSTEM_SYMBOL_TABLE[text].sid)


def _create_local_sid_token(table):
    sids = {}
    for token in table:
        if token.location is None or token.location.name != TEXT_ION:
            sids[token.text] = token.sid

    def token(text):
        return _sid(sids[text])

    return token


def _text_token(text):
    return _tok(text, sid=None)


@listify
@_nextify
def _lst(imports=None, symbols=None, token=_system_sid_token):
    """Generates a list of event pairs for a local symbol table.

    Args:
        sys (SymbolTable): The symbol table to resolve SIDs from.
        imports (Optional[Iterable[_ImportDesc]]): The symbol tables to import.
        symbols (Optional[Iterable[Unicode]]): The local symbols to declare.
        token (Optional[Callable]): Function to construct the token directly from text,
            by default it uses the system symbol table.
    Returns:
        List[Tuple[ReadEvent,IonEvent]]
    """

    yield e_start_struct(annotations=(token(TEXT_ION_SYMBOL_TABLE),))

    if imports is not None:
        if imports is _APPEND:
            yield e_symbol(token(TEXT_ION_SYMBOL_TABLE), field_name=(token(TEXT_IMPORTS)))
        else:
            yield e_start_list(field_name=(token(TEXT_IMPORTS)))
            for desc in imports:
                yield e_start_struct()
                if desc.name is not None:
                    yield e_string(desc.name, field_name=token(TEXT_NAME))
                if desc.version is not None:
                    yield e_int(desc.version, field_name=token(TEXT_VERSION))
                if desc.max_id is not None:
                    yield e_int(desc.max_id, field_name=token(TEXT_MAX_ID))
                yield e_end_struct()
            yield e_end_list()

    if symbols is not None:
        yield e_start_list(field_name=(token(TEXT_SYMBOLS)))
        for symbol_text in symbols:
            yield e_string(symbol_text)
        yield e_end_list()

    yield e_end_struct()


_SHADOW_ION_TEXTS = [
    TEXT_ION_1_0,
    TEXT_ION_SYMBOL_TABLE,
    TEXT_NAME,
    TEXT_VERSION,
    TEXT_MAX_ID,
    TEXT_IMPORTS,
    TEXT_SYMBOLS
]
_SHADOW_ION_TABLE = shared_symbol_table(
    u'shadow_ion',
    1,
    _SHADOW_ION_TEXTS
)
_SHADOW_ION_DESC = _desc(
    _SHADOW_ION_TABLE.name, _SHADOW_ION_TABLE.version, _SHADOW_ION_TABLE.max_id
)


def _test_catalog():
    catalog = SymbolTableCatalog()
    foo_1 = shared_symbol_table(u'foo', 1, [u'a', u'b'])
    foo_3 = shared_symbol_table(u'foo', 3, [u'c', u'd', u'e', u'f'], imports=[foo_1])
    bar_1 = shared_symbol_table(u'bar', 1, [u'x', u'y', u'z'])
    zoo_4 = shared_symbol_table(u'zoo', 4, [u'm', u'n', u'o'])

    for table in [foo_1, foo_3, bar_1, zoo_4, _SHADOW_ION_TABLE]:
        catalog.register(table)

    return catalog


@coroutine
def _predefined_reader(event_pairs):
    expected_iter = (e for e, _ in event_pairs)
    output_iter = iter(add_depths(e for _, e in event_pairs))
    output = None
    while True:
        actual = yield output
        expected = next(expected_iter)
        assert expected == actual
        output = next(output_iter)


class _P(NamedTuple):
    desc: str
    outer: Sequence
    inner: Sequence
    catalog: Any = None

    def __str__(self):
        return self.desc


def _create_lst_params(
        prefix_desc='SYSTEM SIDS',
        prefix_pairs=None,
        token=_system_sid_token,
        append_start=SYSTEM_SYMBOL_TABLE.max_id):
    """

    Args:
        prefix_desc (Optional[String]): The prefix for the test parameter description.
        prefix_pairs (Optional[Iterable[Tuple[DataEvent, IonEvent]]]): The prefix of event pairs to
            put into the inner stream, should only be system values.
        token (Optional[Callable]): The token encoder.
        append_start (Optional[int]): The start of the LST for direct append test cases.
    """
    if prefix_pairs is None:
        prefix_pairs = [(NEXT, IVM)]
    params = [
        _P(
            desc='LOCAL ONLY',
            inner=_lst(symbols=[u'a', u'b', u'c'], token=token) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(11))),
                (NEXT, e_symbol(_sid(12))),
                (NEXT, e_symbol(None)),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', 10))),
                (NEXT, e_symbol(_tok(u'b', 11))),
                (NEXT, e_symbol(_tok(u'c', 12))),
                (NEXT, e_symbol(None)),
            ],
        ),
        _P(
            desc='LOCAL ONLY RESET',
            inner=_lst(symbols=[u'a'], token=token) + [
                (NEXT, e_symbol(_sid(10))),
            ] + _lst(symbols=[u'b'], token=_system_sid_token) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, IVM),
                (NEXT, e_symbol(_sid(10))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', 10))),
                (NEXT, e_symbol(_tok(u'b', 10))),
                (NEXT, IonException),
            ],
        ),
        _P(
            desc='$ion_1_0 USED AS NOP',
            inner=_lst(symbols=[u'$ion_1_0', u'a'], token=token) + [
                (NEXT, e_symbol(_sid(11))),
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(11))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', 11))),
                (NEXT, e_symbol(_tok(u'a', 11))),
            ],
        ),
        _P(
            desc='IMPORTS ONLY - PLACEHOLDER',
            inner=_lst(imports=[_desc(u'unknown', 1, 10), _desc(u'other', 1, 5)], token=token) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(22))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(None, 10, _loc(u'unknown', 1)))),
                (NEXT, e_symbol(_tok(None, 22, _loc(u'other', 3)))),
            ],
        ),
        _P(
            desc='IMPORTS ONLY - EXACT MATCH, NO MAX ID',
            inner=_lst(imports=[_desc(u'foo', 3), _desc(u'bar', 1)], token=token) + [
                (NEXT, e_symbol(_sid(15))),
                (NEXT, e_symbol(_sid(16))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'f', 15, _loc(u'foo', 6)))),
                (NEXT, e_symbol(_tok(u'x', 16, _loc(u'bar', 1)))),
            ],
        ),
        _P(
            desc='IMPORTS ONLY - EXACT MATCH, NO VERSION, NO MAX ID',
            inner=_lst(imports=[_desc(u'foo'), _desc(u'bar')], token=token) + [
                (NEXT, e_symbol(_sid(11))),
                (NEXT, e_symbol(_sid(13))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'b', 11, _loc(u'foo', 2)))),
                (NEXT, e_symbol(_tok(u'y', 13, _loc(u'bar', 2)))),
            ],
        ),
        _P(
            desc='IMPORTS ONLY - NO MATCH, NO VERSION, NO MAX ID',
            inner=_lst(imports=[_desc(u'zoo')], token=token),
            outer=[
                (NEXT, CannotSubstituteTable),
            ],
        ),
        _P(
            desc='IMPORTS ONLY - INEXACT MATCH, NO VERSION',
            inner=_lst(imports=[_desc(u'bar', max_id=2)], token=token) + [
                (NEXT, e_symbol(_sid(11))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'y', 11, _loc(u'bar', 2)))),
            ],
        ),
        _P(
            desc='IMPORTS AND LOCALS',
            inner=_lst(
                imports=[_desc(u'foo', 3, 4)],
                symbols=[u'aa', u'bb', u'cc', u'dd'],
                token=token
            ) + [
                (NEXT, e_symbol(_sid(13))),
                (NEXT, e_symbol(_sid(15))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'd', 13, _loc(u'foo', 4)))),
                (NEXT, e_symbol(_tok(u'bb', 15))),
            ],
        ),
        _P(
            desc='APPEND SYSTEM SYMBOLS',
            inner=_lst(imports=_APPEND, symbols=[u'a', u'b', u'c'], token=token) + [
                (NEXT, e_symbol(_sid(append_start + 1))),
                (NEXT, e_symbol(_sid(append_start + 3))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', append_start + 1))),
                (NEXT, e_symbol(_tok(u'c', append_start + 3))),
            ],
        ),
        _P(
            desc='APPEND TO PREVIOUS LST',
            inner=_lst(symbols=[u'a', u'b', u'c'], token=token) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(12))),
            ] + _lst(imports=_APPEND, symbols=[u'd', u'e', u'f'], token=_system_sid_token) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(13))),
                (NEXT, e_symbol(_sid(15))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', 10))),
                (NEXT, e_symbol(_tok(u'c', 12))),
                (NEXT, e_symbol(_tok(u'a', 10))),
                (NEXT, e_symbol(_tok(u'd', 13))),
                (NEXT, e_symbol(_tok(u'f', 15))),
            ],
        ),
        _P(
            desc='APPEND TO PREVIOUS LST WITH IMPORTS',
            inner=_lst(
                imports=[_desc(u'foo', 3, 4)],
                symbols=[u'aa', u'bb', u'cc'],
                token=token
            ) + [
                (NEXT, e_symbol(_sid(10))),
                (NEXT, e_symbol(_sid(14))),
            ] + _lst(imports=_APPEND, symbols=[u'dd', u'ee', u'ff'], token=_system_sid_token) + [
                (NEXT, e_symbol(_sid(11))),
                (NEXT, e_symbol(_sid(15))),
                (NEXT, e_symbol(_sid(18))),
            ],
            outer=[
                (NEXT, e_symbol(_tok(u'a', 10, _loc(u'foo', 1)))),
                (NEXT, e_symbol(_tok(u'aa', 14))),
                (NEXT, e_symbol(_tok(u'b', 11, _loc(u'foo', 2)))),
                (NEXT, e_symbol(_tok(u'bb', 15))),
                (NEXT, e_symbol(_tok(u'ee', 18))),
            ],
        ),
    ]
    for param in params:
        yield _P(
            desc='LST - ' + prefix_desc + ' - ' + param.desc,
            catalog=_test_catalog(),
            inner= prefix_pairs + param.inner,
            outer=param.outer
        )


_BASIC_PARAMS = [
    _P(
        desc='IVM ONLY',
        catalog=None,
        inner=[
            (NEXT, IVM),
            (NEXT, IVM),
            (NEXT, END),
        ],
        outer=[
            (NEXT, END),
        ],
    ),
    _P(
        desc='SYSTEM SYMBOL',
        catalog=None,
        inner=[
            (NEXT, IVM),
            (NEXT, e_symbol(_sid(4))),
        ],
        outer=[
            (NEXT, e_symbol(_SYS[4])),
        ],
    ),
    _P(
        desc='SYSTEM ANNOTATION',
        catalog=None,
        inner=[
            (NEXT, IVM),
            (NEXT, e_int(100, annotations=(_sid(5),))),
        ],
        outer=[
            (NEXT, e_int(100, annotations=(_SYS[5],))),
        ],
    ),
    _P(
        desc='SYSTEM FIELD_NAME',
        catalog=None,
        inner=[
            (NEXT, IVM),
            # We can cheat because we don't have a real underlying reader.
            (NEXT, e_string(u'hi', field_name=_sid(6))),
        ],
        outer=[
            (NEXT, e_string(u'hi', field_name=_SYS[6])),
        ],
    ),
]


@parametrize(*chain(
    _BASIC_PARAMS,
    _create_lst_params(),
    _create_lst_params(
        prefix_desc='SHADOW IMPORT',
        prefix_pairs=_lst(imports=[_SHADOW_ION_DESC]),
        token=_create_local_sid_token(local_symbol_table([_SHADOW_ION_TABLE])),
        append_start=SYSTEM_SYMBOL_TABLE.max_id + _SHADOW_ION_TABLE.max_id
    ),
    _create_lst_params(
        prefix_desc='SHADOW LOCAL',
        prefix_pairs=_lst(symbols=_SHADOW_ION_TEXTS),
        token=_create_local_sid_token(local_symbol_table(symbols=_SHADOW_ION_TEXTS)),
        append_start=SYSTEM_SYMBOL_TABLE.max_id + len(_SHADOW_ION_TEXTS)
    ),
    _create_lst_params(
        prefix_desc='TEXT TOKENS',
        token=_text_token,
    ),
))
def test_managed_reader(p):
    reader_scaffold(managed_reader(_predefined_reader(p.inner), p.catalog), p.outer)


def test_managed_thunk_event():
    event = IonEvent(
        IonEventType.SCALAR, IonType.INT, 5, _tok(u'foo', None), ()
    )
    thunk_event = _IonManagedThunkEvent(
        IonEventType.SCALAR, IonType.INT, lambda: 5, lambda: _tok(u'foo', None), lambda: (), None
    )

    assert thunk_event == event

    assert thunk_event.derive_annotations((_tok(u'bar', None),)) \
        == event.derive_annotations((_tok(u'bar', None),))

    assert thunk_event.derive_annotations(lambda: (_tok(u'bar', None),)) \
        == event.derive_annotations((_tok(u'bar', None),))

    assert thunk_event.derive_field_name(_tok(u'bar', None)) \
        == event.derive_field_name(_tok(u'bar', None))

    assert thunk_event.derive_field_name(lambda: _tok(u'bar', None)) \
        == event.derive_field_name(_tok(u'bar', None))



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import sys
from setuptools import setup, find_packages, Extension
from install import _install_ionc

C_EXT = True if not hasattr(sys, 'pypy_translation_info') else False


def run_setup():
    if C_EXT and _install_ionc():
        print('C extension is enabled!')
        kw = dict(
            ext_modules=[
                Extension(
                    'amazon.ion.ionc',
                    sources=['amazon/ion/ioncmodule.c'],
                    include_dirs=['amazon/ion/ion-c-build/include',
                                  'amazon/ion/ion-c-build/include/ionc',
                                  'amazon/ion/ion-c-build/include/decNumber'],
                    libraries=['ionc', 'decNumber'],
                    library_dirs=['amazon/ion/ion-c-build/lib'],
                    extra_link_args=['-Wl,-rpath,%s' % '$ORIGIN/ion-c-build/lib',  # LINUX
                                     '-Wl,-rpath,%s' % '@loader_path/ion-c-build/lib'  # MAC
                                     ],
                    extra_compile_args=['-std=c99'],
                ),
            ],
        )
    else:
        print('Using pure python implementation.')
        kw = dict()


    setup(
        name='amazon.ion',
        version='0.11.2',
        description='A Python implementation of Amazon Ion.',
        url='http://github.com/amazon-ion/ion-python',
        author='Amazon Ion Team',
        author_email='ion-team@amazon.com',
        license='Apache License 2.0',


        packages=find_packages(exclude=['tests*']),
        include_package_data=True,
        namespace_packages=['amazon'],

        setup_requires=[
            'pytest-runner',
        ],

        tests_require=[
            'pytest',
        ],
        **kw
    )


run_setup()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from functools import partial

from datetime import timedelta, datetime
from io import BytesIO
from itertools import chain

from decimal import Decimal
from typing import NamedTuple

from amazon.ion.core import OffsetTZInfo, IonEvent, IonType, IonEventType
from amazon.ion.exceptions import IonException

from tests import parametrize

from amazon.ion.simpleion import loads, dumps
from amazon.ion.symbols import SymbolToken
from amazon.ion.writer import blocking_writer
from amazon.ion.writer_text import raw_writer
from tests.writer_util import assert_writer_events, WriterParameter, generate_scalars, generate_containers, \
    SIMPLE_SCALARS_MAP_TEXT

_D = Decimal
_DT = datetime

_E = IonEvent
_IT = IonType
_ET = IonEventType
_P = WriterParameter

_EMPTY_CONTAINER_MAP = {
    _IT.LIST: (
        (
            (),
            b'[]',
        ),
    ),
    _IT.SEXP: (
        (
            (),
            b'()',
        ),
    ),
    _IT.STRUCT: (
        (
            (),
            b'{}',
        ),
    ),
}

_generate_simple_scalars = partial(generate_scalars, SIMPLE_SCALARS_MAP_TEXT)
_generate_empty_containers = partial(generate_containers, _EMPTY_CONTAINER_MAP)

_SIMPLE_ANNOTATIONS = (
    SymbolToken(None, 4), # System symbol 'name'.
    u'\x00',
    u'\uff4e', # An full-width latin 'n' code point.
    u'\U0001f4a9', # A 'pile of poo' emoji code point.
)
_SIMPLE_ANNOTATIONS_ENCODED = br"$4::'\x00'::'\uff4e'::'\U0001f4a9'::"


def _generate_annotated_values():
    for value_p in chain(_generate_simple_scalars(), _generate_empty_containers()):
        events = (value_p.events[0].derive_annotations(_SIMPLE_ANNOTATIONS),) + value_p.events[1:]

        if isinstance(value_p.expected, (tuple, list)):
            expecteds = value_p.expected
        else:
            expecteds = (value_p.expected,)
        final_expecteds = ()
        for expected in expecteds:
            final_expected = _SIMPLE_ANNOTATIONS_ENCODED + expected
            final_expecteds += (final_expected, )
        yield _P(
            desc='ANN %s' % value_p.desc,
            events=events,
            expected=final_expecteds,
        )


_SIMPLE_FIELD_NAME = u'field'
_TOKEN_FIELD_NAME = SymbolToken(None, 4) # System symbol 'name'.


def _generate_simple_containers(*generators, **opts):
    """Composes the empty tests with the simple scalars to make singletons."""
    repeat = opts.get('repeat', 1)
    targets = tuple(chain(*generators))
    for empty_p in _generate_empty_containers():
        start_event, end_event = empty_p.events
        delim = b','
        if start_event.ion_type is _IT.SEXP:
            delim = b' '
        for value_p in targets:
            field_names = (None,)
            if start_event.ion_type is _IT.STRUCT:
                field_names = (_SIMPLE_FIELD_NAME, _TOKEN_FIELD_NAME)
            for field_name in field_names:
                b_start = empty_p.expected[0:1]
                b_end = empty_p.expected[-1:]

                value_event = value_p.events[0]
                if isinstance(value_p.expected, (tuple, list)):
                    value_expected = value_p.expected[0]
                else:
                    value_expected = value_p.expected
                if field_name is not None:
                    value_event = value_event.derive_field_name(field_name)
                    if isinstance(field_name, SymbolToken):
                        value_expected = \
                            (u'$%d:'% field_name.sid).encode() + value_expected
                    else:
                        value_expected = \
                            field_name.encode() + b":" + value_expected
                # Make sure to compose the rest of the value (if not scalar, i.e. empty containers).
                value_events = (value_event,) + value_p.events[1:]
                events = (start_event,) + (value_events * repeat) + (end_event,)

                expected = b_start + delim.join([value_expected] * repeat) + b_end

                yield _P(
                    desc='SINGLETON %s %r' % (start_event.ion_type.name, expected),
                    events=events,
                    expected=expected
                )


_P_TOP_LEVEL = [
    _P(
        desc='TOP-LEVEL IVM',
        events=[
            _E(_ET.VERSION_MARKER),
        ],
        expected=b'$ion_1_0',
    ),
    _P(
        desc='TOP-LEVEL IVM x2',
        events=[
            _E(_ET.VERSION_MARKER),
            _E(_ET.VERSION_MARKER),
            _E(_ET.STREAM_END),
        ],
        expected=b'$ion_1_0 $ion_1_0',
    ),
    _P(
        desc='TOP-LEVEL STREAM END',
        events=[
            _E(_ET.STREAM_END),
        ],
        expected=b'',
    ),
    _P(
        desc='TOP-LEVEL INCOMPLETE',
        events=[
            _E(_ET.INCOMPLETE),
        ],
        expected=TypeError,
    ),
    _P(
        desc='TOP-LEVEL CONTAINER_END',
        events=[
            _E(_ET.CONTAINER_END, _IT.LIST),
        ],
        expected=TypeError,
    ),
    _P(
        desc="NULL WITH VALUE",
        events=[
            _E(_ET.SCALAR, _IT.NULL, u'foo')
        ],
        expected=TypeError
    ),
]


def new_writer():
    buf = BytesIO()
    return buf, blocking_writer(raw_writer(), buf)


@parametrize(
    *tuple(chain(
        _P_TOP_LEVEL,
        _generate_simple_scalars(),
        _generate_empty_containers(),
        _generate_annotated_values(),
        _generate_simple_containers(
            _generate_simple_scalars(),
            _generate_empty_containers(),
            _generate_annotated_values(),
            # Add an extra level of depth.
            _generate_simple_containers(
                _generate_empty_containers(),
            )
        ),
        _generate_simple_containers(
            _generate_annotated_values(),
            repeat=4
        ),
    ))
)
def test_raw_writer(p):
    assert_writer_events(p, new_writer)

@parametrize(
        (None, True),
        ('', True),
        (' ', True),
        (' ', True),
        ('\t', True),
        ('\n', True),
        ('       ', True),
        ('  \t     \n', True),
        ('  \t  x  \n', False),
        ('\"', False),
        ('apple', False),
        ('1', False))
def test_valid_indent_strs(p):
    indent, is_valid = p
    try:
        raw_writer(indent)
        assert is_valid
    except ValueError as e:
        assert not is_valid
        assert 'only whitespace' in str(e)
    if is_valid:
        # we only allow indent strings to be used to be used if they produce valid ion.
        ion_val = loads('[a, {x:2, y: height::17}]')
        assert ion_val == loads(dumps(ion_val, binary=False, indent=indent))

class _P_Q(NamedTuple):
    symbol: str
    needs_quotes: bool
    backslash_required: bool = False

    pass

@parametrize(
        _P_Q('a', False),
        _P_Q('a0', False),
        _P_Q('_a_0_', False),
        _P_Q('int', False),
        _P_Q('int32', False),
        _P_Q('FALSE', False),
        _P_Q('False', False),
        _P_Q('bool', False),
        _P_Q('NaN', False),
        _P_Q('T', False),
        _P_Q('type', False),
        _P_Q('inf', False),
        # begin needs quotes
        _P_Q('a.0', True),
        _P_Q('1', True),
        _P_Q('{', True),
        _P_Q('\'', True, backslash_required=True),
        _P_Q('"', True),
        _P_Q('\\', True, backslash_required=True),
        # begin reserved words
        _P_Q('nan', True),
        _P_Q('null', True),
        _P_Q('null.int', True),
        _P_Q('+inf', True),
        _P_Q('-inf', True),
        _P_Q('false', True),
        _P_Q('true', True),
        )
def test_quote_symbols(p):
    symbol_text, needs_quotes, backslash_required = p

    try:
        loads(symbol_text + '::4')
        threw_without_quotes = False
    except IonException as e:
        threw_without_quotes = True

    quoted_symbol_text = "'" + ('\\' if backslash_required else '') + symbol_text + "'"
    ion_value = loads(quoted_symbol_text + "::4")
    quoted = "'" in dumps(ion_value, binary=False)

    assert needs_quotes == threw_without_quotes
    assert needs_quotes == quoted
    assert len(ion_value.ion_annotations) == 1
    assert ion_value.ion_annotations[0].text == symbol_text
